from django.test import TestCase
from django.contrib.auth.models import Group
from django.contrib.auth.models import Permission
from rest_framework.test import APIClient
from rest_framework.reverse import reverse

from munch.apps.users.tests.factories import UserFactory

from ...models import Contact
from ...models import ContactList
from ...models import ContactQueue
from ...models import CollectedContact


class ContactListAPITestCase(TestCase):
    def setUp(self):
        self.api_version = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_contact_list_default_contact_fields(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 2',
                'contact_fields': [
                    {
                        'name': 'beard_color',
                        'type': 'Char',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(0, response.data['contacts_count'])
        contact_list = ContactList.objects.all().first()
        self.assertEqual(len(contact_list.contact_fields), 1)
        self.assertEqual(contact_list.contact_fields[0].get('type'), 'Char')
        self.assertEqual(contact_list.contact_fields[0].get('required'), True)
        self.assertEqual(
            contact_list.contact_fields[0].get('name'), 'beard_color')

    def test_contact_list_default_empty_contact_fields(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 2',
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(0, response.data['contacts_count'])
        contact_list = ContactList.objects.all().first()
        self.assertEqual(
            len(contact_list.contact_fields),
            len(ContactList.DEFAULT_CONTACT_FIELDS))
        for default_field in ContactList.DEFAULT_CONTACT_FIELDS:
            self.assertIn(
                default_field.get('name'),
                [f.get('name') for f in contact_list.contact_fields])

    def test_contact_list_creation(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'has_beard',
                        'type': 'Boolean',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'},
            format='json')

        self.assertEqual(201, response.status_code)

        url = response.data['url']

        response = self.client.get(url)
        self.assertEqual(200, response.status_code)
        self.assertEqual(0, response.data['contacts_count'])

        response = self.client.get(response.data['_links']['contacts']['href'])

        self.assertEqual(200, response.status_code)
        self.assertEqual(0, response.data['count'])

        self.client.logout()
        another_user = UserFactory(groups=['managers'])
        self.client.login(
            identifier=another_user.identifier, password='password')

        response = self.client.get(url)
        self.assertEqual(403, response.status_code)

    def test_contact_list_double_creation(self):
        """ Same-named contact list """
        ContactList.objects.create(name='Name', author=self.user)

        # same name
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version), {'name': 'Name'})
        self.assertEqual(response.status_code, 400)

    def test_contact_creation(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'has_beard',
                        'type': 'Boolean',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'},
            format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'has_beard': True}},
            format='json')

        self.assertEqual(201, response.status_code)

        response = self.client.get(url)

        self.assertEqual(1, response.data['contacts_count'])
        mails_response = self.client.get(url + 'contacts/')
        self.assertEqual(mails_response.data['count'], 1)

        response = self.client.get(response.data['_links']['contacts']['href'])

        self.assertEqual(200, response.status_code)
        self.assertEqual(1, response.data['count'])

        # another list don't get the contacts

        response_list2 = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 2',
                'contact_fields': [
                    {
                        'name': 'has_beard',
                        'type': 'Boolean',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'},
            format='json')

        url_list2 = response_list2.data['url']
        list2 = self.client.get(url_list2).data
        self.assertEqual(list2['contacts_count'], 0)

        list2_mails = self.client.get(url_list2 + 'contacts/').data
        self.assertEqual(list2_mails['count'], 0)

    def test_bulk_contact_creation(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'has_beard',
                        'type': 'Boolean',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'},
            format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            data=[
                {
                    'address': 'test1@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': True}},
                {
                    'address': 'test2@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}},
                {
                    'address': 'test3@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}}],
            format='json')

        self.assertEqual(201, response.status_code)

        response = self.client.get(url)

        self.assertEqual(3, response.data['contacts_count'])

    def test_csv_contact_creation(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'has_beard',
                        'type': 'Boolean',
                        'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'},
            format='json')

        list_url = response.data['url']
        data = [
            {'address': 'test1@example.com', 'properties.has_beard': True},
            {'address': 'test2@example.com', 'properties.has_beard': False},
            {'address': 'test3@example.com', 'properties.has_beard': False}]
        response = self.client.post(
            list_url + 'contacts/', data=data, format='csv')

        self.assertEqual(201, response.status_code)

        response = self.client.get(list_url)

        self.assertEqual(3, response.data['contacts_count'])

        mails_response = self.client.get(list_url + 'contacts/')

        self.assertEqual(len(data), len(mails_response.data['results']))
        # Check if each address are in mails_response.data['results']
        self.assertEqual(
            sorted([c.get('address') for c in data]),
            sorted([c.get('address') for c in mails_response.data['results']]))
        for contact in data:
            # Retrieve contact from mails_response.data['results']
            # based on address
            data_contact = [c for c in mails_response.data['results'] if c.get(
                'address') == contact.get('address')][0]
            self.assertEqual(
                data_contact.get(
                    'properties')['has_beard'], contact.get(
                        'properties.has_beard'))

    def test_filter_properties(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'has_beard',
                        'type': 'Boolean',
                        'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'},
            format='json')

        list_url = response.data['url']

        response = self.client.post(
            list_url + 'contacts/?fields=[has_beard,has_mustache]',
            data=[{'address': 'test1@example.com',
                   'has_beard': 'True',
                   'has_mustache': 'False',
                   'is_hot': 'dunno'}],
            format='csv'
        )
        self.assertEqual(201, response.status_code)

        mails_response = self.client.get(list_url + 'contacts/')
        first_item = mails_response.data['results'][0]
        self.assertEqual(first_item['address'], 'test1@example.com')
        # 'is_hot' should not be included
        self.assertEqual(
            set(first_item['properties']), set(['has_beard']))

    def test_contact_field_add_empty_name(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': '',
                        'type': 'DateTime',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')
        self.assertEqual(400, response.status_code)

        response = self.client.post(
            '/v1/contacts/lists/',
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': '  ',
                        'type': 'Char',
                        'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')
        self.assertEqual(400, response.status_code)

    def test_contact_field_add_duplicate_field_name(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True},
                    {
                        'name': 'beArd_birtH',
                        'type': 'DateTime',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(len(response.data['contact_fields']), 1)
        self.assertIn('beard_birth', response.data['contact_fields'])

        response = self.client.post(
            '/v1/contacts/lists/',
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'Char',
                        'required': True},
                    {
                        'name': 'beard_birth',
                        'type': 'Char',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(len(response.data['contact_fields']), 1)
        self.assertIn('beard_birth', response.data['contact_fields'])

    def test_contact_field_add_non_required_field(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_birth',
                    'type': 'DateTime',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-12-31T10:30:33.000000Z'}},
            format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True
                    }, {
                        'name': 'beard_size',
                        'type': 'Float',
                        'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('contact_fields')), 2)

    def test_contact_field_add_required_field(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_birth',
                    'type': 'DateTime',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-12-31T10:30:33.000000Z'}},
            format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True
                    }, {
                        'name': 'beard_size',
                        'type': 'Float',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            len(response.data.get('contact_fields').get('beard_size')), 1)

    def test_contact_field_add_required_field_trimed(self):
        response = self.client.post(
            '/v1/contacts/lists/',
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_name',
                        'type': 'Char',
                        'required': True
                    }, {
                        'name': 'beard_subname',
                        'type': 'Char',
                        'required': False
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/v1/contacts/',
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_name': '     ', 'beard_subname': '  '}},
            format='json')

        self.assertEqual(400, response.status_code)
        self.assertEqual(Contact.objects.count(), 0)

        response = self.client.post(
            '/v1/contacts/',
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {
                    'beard_name': 'My Beard', 'beard_subname': '  '}},
            format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(
            ContactList.objects.all().first().properties.get(
                'beard_subname'), None)

    def test_contact_field_set_existing_field_as_required(self):
        """
            Set a contact_field to required with contact that
            doesn't have value setted. Then, this change can't be done.
        """
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_birth',
                    'type': 'DateTime',
                    'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {}},
            format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test2@example.com',
                'contact_list': url,
                'properties': {"beard_birth": "2015-12-31T10:30:33.000000Z"}},
            format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 2)

        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            len(response.data.get('contact_fields').get('beard_birth')), 1)
        self.assertEqual(Contact.objects.count(), 2)

    def test_contact_field_set_existing_field_as_required_not_empty(self):
        """
            Set a contact_field to required with all contacts with
            a value setted. Then, this change can be done.
        """
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_birth',
                    'type': 'DateTime',
                    'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-12-31T10:30:33.000000Z'}},
            format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('contact_fields')), 1)
        self.assertTrue(response.data.get('contact_fields')[0].get('required'))
        self.assertEqual(
            response.data.get('contact_fields')[0].get('name'), 'beard_birth')
        self.assertEqual(Contact.objects.count(), 1)

    def test_contact_field_remove_field(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True},
                    {
                        'name': 'beard_size',
                        'type': 'Float',
                        'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {
                    'beard_birth': '2015-12-31T10:30:33.000000Z',
                    'beard_size': 12}},
            format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('contact_fields')), 1)
        self.assertEqual(
            response.data.get('contact_fields')[0].get('name'), 'beard_birth')
        self.assertIsNone(Contact.objects.first().properties.get('beard_size'))
        self.assertIsNotNone(
            Contact.objects.first().properties.get('beard_birth'))

    def test_contact_field_remove_multiple_fields(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True},
                    {
                        'name': 'beard_size',
                        'type': 'Float',
                        'required': True},
                    {
                        'name': 'beard_color',
                        'type': 'Char',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {
                    'beard_birth': '2015-12-31T10:30:33.000000Z',
                    'beard_color': 'Blue',
                    'beard_size': 12}},
            format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('contact_fields')), 1)
        self.assertEqual(
            response.data.get('contact_fields')[0].get('name'), 'beard_birth')
        self.assertIsNone(Contact.objects.first().properties.get('beard_size'))
        self.assertIsNotNone(
            Contact.objects.first().properties.get('beard_birth'))

    def test_contact_field_change_type_to_char(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'DateTime',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-12-31T10:30:33.000000Z'}},
            format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_birth',
                        'type': 'Char',
                        'required': True
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('contact_fields')), 1)
        self.assertEqual(
            response.data.get('contact_fields')[0].get('name'), 'beard_birth')
        self.assertEqual(
            response.data.get('contact_fields')[0].get('type'), 'Char')
        self.assertIsNotNone(
            Contact.objects.first().properties.get('beard_birth'))

    def test_contact_field_change_type_to_non_char(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_color',
                        'type': 'Char',
                        'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_color': 'Blue'}},
            format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        contact_list = ContactList.objects.all().first()

        response = self.client.put(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk),
            {
                'name': 'Test list 1',
                'contact_fields': [
                    {
                        'name': 'beard_color',
                        'type': 'Integer',
                        'required': True
                    }],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            len(response.data.get('contact_fields').get('beard_color')), 1)
        self.assertEqual(Contact.objects.count(), 1)

        response = self.client.get(
            '/{}/contacts/lists/{}/'.format(self.api_version, contact_list.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('contact_fields')), 1)
        self.assertEqual(
            response.data.get('contact_fields')[0].get('name'), 'beard_color')
        self.assertEqual(
            response.data.get('contact_fields')[0].get('type'), 'Char')


class ContactPropertiesAPITestCase(TestCase):
    def setUp(self):
        self.api_version = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_contact_field_dropped(self):
        """ Fields not in contact_fields must be dropped """
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {
                    'has_mustache': True,
                    'has_beard': False}
            }, format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        self.assertNotIn(
            'has_mustache', Contact.objects.first().properties.keys())

    def test_contact_field_required(self):
        """ Required fields must raise ValidationError if not present """
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'has_mustache': True}}, format='json')

        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.data.get('properties').get('has_beard')), 1)
        self.assertEqual(Contact.objects.count(), 0)

    def test_contact_field_boolean_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Char',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'has_beard': True}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

    def test_contact_field_string_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Char',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'has_beard': True}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

    def test_contact_field_integer_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Integer',
                    'required': True}, {
                    'name': 'has_mustache',
                    'type': 'Integer',
                    'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'has_beard': 12}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

    def test_contact_field_wrong_integer_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Integer',
                    'required': True}, {
                    'name': 'has_mustache',
                    'type': 'Integer',
                    'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']
        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'has_beard': '12aa'}}, format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.data.get('properties').get('has_beard')), 1)
        self.assertEqual(Contact.objects.count(), 0)

    def test_contact_field_float_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_size',
                    'type': 'Float',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_size': 1.65}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

    def test_contact_field_wrong_float_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_size',
                    'type': 'Float',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_size': True}}, format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.data.get('properties').get('beard_size')), 1)
        self.assertEqual(Contact.objects.count(), 0)

    def test_contact_field_date_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_age',
                    'type': 'Date',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_age': '2015-12-31'}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

    def test_contact_field_wrong_date_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_age',
                    'type': 'Date',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_age': '2015-003-31'}}, format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.data.get('properties').get('beard_age')), 1)
        self.assertEqual(Contact.objects.count(), 0)

    def test_contact_field_datetime_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_birth',
                    'type': 'DateTime',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-12-31T10:30:33.000000Z'}},
            format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test2@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-12-31 10:30:33'}},
            format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 2)

    def test_contact_field_wrong_datetime_type(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'beard_birth',
                    'type': 'DateTime',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {'beard_birth': '2015-13-31'}},
            format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.data.get('properties').get('beard_birth')), 1)
        self.assertEqual(Contact.objects.count(), 0)

    def test_contact_field_duplicate_name(self):
        response = self.client.post(
            '/v1/contacts/lists/',
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Char',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/v1/contacts/',
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {
                    'has_beard': 'test',
                    'has_beard ': 'test2'}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        self.assertEqual(
            Contact.objects.all().first().properties, {'has_beard': 'test'})

    def test_contact_field_add_empty_property(self):
        response = self.client.post(
            '/v1/contacts/lists/',
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Char',
                    'required': False}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/v1/contacts/',
            {
                'address': 'test@example.com',
                'contact_list': url,
                'properties': {
                    '     ': 'test2',
                    'has_beard': 'test'}}, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 1)
        self.assertEqual(
            Contact.objects.all().first().properties, {'has_beard': 'test'})


class CollectedContactPropertiesAPITestCase(TestCase):
    def setUp(self):
        self.api_version = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_contact_field_dropped(self):
        """ Fields not in contact_fields must be dropped """
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'has_beard', 'type': 'Boolean', 'required': True}])
        q.save()
        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'has_beard': True,
                'has_mustache': True})

        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)
        self.assertNotIn(
            'has_mustache', CollectedContact.objects.first().properties.keys())

    def test_contact_field_required(self):
        """ Required fields must raise ValidationError if not present """
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'has_beard', 'type': 'Boolean', 'required': True}])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'has_mustache': True})

        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.context[0].get('errors').get('has_beard')), 1)
        self.assertEqual(CollectedContact.objects.count(), 0)

    def test_contact_field_boolean_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'has_beard', 'type': 'Char', 'required': True}])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'has_beard': True})
        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)

    def test_contact_field_string_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'has_beard', 'type': 'Char', 'required': True}])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'has_beard': True})

        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)

    def test_contact_field_integer_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'has_beard', 'type': 'Integer', 'required': True},
                {'name': 'has_mustache', 'type': 'Integer', 'required': False}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'has_beard': 12})

        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)

    def test_contact_field_wrong_integer_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'has_beard', 'type': 'Integer', 'required': True},
                {'name': 'has_mustache', 'type': 'Integer', 'required': False}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'has_beard': '12aa'})

        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.context[0].get('errors').get('has_beard')), 1)
        self.assertEqual(CollectedContact.objects.count(), 0)

    def test_contact_field_float_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'beard_size', 'type': 'Float', 'required': True}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'beard_size': 1.65})

        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)

    def test_contact_field_wrong_float_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'beard_size', 'type': 'Float', 'required': True}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'beard_size': True})

        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.context[0].get('errors').get('beard_size')), 1)
        self.assertEqual(CollectedContact.objects.count(), 0)

    def test_contact_field_date_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'beard_age', 'type': 'Date', 'required': True}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'beard_age': '2015-12-31'})
        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)

    def test_contact_field_wrong_date_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'beard_age', 'type': 'Date', 'required': True}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'beard_age': '2015-003-31'})

        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.context[0].get('errors').get('beard_age')), 1)
        self.assertEqual(CollectedContact.objects.count(), 0)

    def test_contact_field_datetime_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'beard_birth', 'type': 'DateTime', 'required': True}
            ])
        q.save()

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope@example.org',
                'beard_birth': '2015-12-31T10:30:33.000000Z'})

        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 1)

        response = self.client.post(
            reverse(
                'subscription',
                kwargs={'uuid': q.uuid}),
            {
                'address': 'nope2@example.org',
                'beard_birth': '2015-12-31 10:30:33'})

        self.assertEqual(201, response.status_code)
        self.assertEqual(CollectedContact.objects.count(), 2)

    def test_contact_field_wrong_datetime_type(self):
        q = ContactQueue(
            author=self.user,
            contact_fields=[
                {'name': 'beard_birth', 'type': 'DateTime', 'required': True}])
        q.save()

        response = self.client.post(
            reverse('subscription', kwargs={'uuid': q.uuid}),
            {'address': 'nope@example.org', 'beard_birth': '2015-13-31'})

        self.assertEqual(400, response.status_code)
        self.assertEqual(
            len(response.context[0].get('errors').get('beard_birth')), 1)
        self.assertEqual(CollectedContact.objects.count(), 0)


class ListMergeTest(TestCase):
    def setUp(self):
        self.api_version = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

        self.cl1 = ContactList.objects.create(
            name='self.cl-1', author=self.user)
        self.cl2 = ContactList.objects.create(
            name='self.cl-2', author=self.user)
        self.cl3 = ContactList.objects.create(
            name='self.cl-3', author=self.user)

        Contact.objects.create(address='1@example.com', contact_list=self.cl1)
        Contact.objects.create(address='2@example.com', contact_list=self.cl2)
        Contact.objects.create(address='3@example.com', contact_list=self.cl3)

    def test_contact_list_merge(self):
        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk),
            [
                'http://testserver' + reverse(
                    '{}:contacts:contactlist-detail'.format(self.api_version),
                    kwargs={'pk': self.cl2.pk}),
                'http://testserver' + reverse(
                    '{}:contacts:contactlist-detail'.format(self.api_version),
                    kwargs={'pk': self.cl3.pk})
            ], format='json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.cl1.contacts.count(), 3)
        self.assertFalse(ContactList.objects.filter(pk=self.cl2.pk).exists())
        self.assertFalse(ContactList.objects.filter(pk=self.cl3.pk).exists())

    def test_contact_list_merge_bad_list(self):
        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk), ['http://example.com/foo/bar'],
            format='json')
        self.assertEqual(response.status_code, 400)

    def test_contact_list_merge_not_my_list(self):
        self.another_user = UserFactory(groups=['users'])
        self.client.login(
            identifier=self.another_user.identifier, password='password')
        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk), [
                    'http://testserver' + reverse(
                        '{}:contacts:contactlist-detail'.format(
                            self.api_version),
                        kwargs={'pk': self.cl2.pk})],
            format='json')
        self.assertEqual(response.status_code, 400)

    def test_contact_list_merge_no_list(self):
        self.another_user = UserFactory(groups=['users'])
        self.client.login(
            identifier=self.another_user.identifier, password='password')
        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk), [], format='json')
        self.assertEqual(response.status_code, 200)

    def test_contact_list_merge_with_itself(self):
        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk),
            ['/{}/contacts/lists/{}/'.format(self.api_version, self.cl1.pk)],
            format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.cl1.contacts.count(), 1)

    def test_contact_list_merge_no_delete_perm(self):
        # alter permissions to remove the right to DELETE lists
        moderators = Group.objects.get(name='managers')

        delete_perms = Permission.objects.filter(
            codename__startswith='delete_', codename__endswith='_contactlist')
        for perm in delete_perms:
            moderators.permissions.remove(perm)
        self.client.logout()
        self.client.login(identifier=self.user.identifier, password='password')

        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk),
            [
                'http://testserver' + reverse(
                    '{}:contacts:contactlist-detail'.format(self.api_version),
                    kwargs={'pk': self.cl2.pk})], format='json')
        self.assertEqual(response.status_code, 400)

    def test_contact_list_merge_doubles(self):
        # It already exists in self.cl1
        double_contact = Contact.objects.create(
            address='1@example.com', contact_list=self.cl2)
        response = self.client.post(
            '/{}/contacts/lists/{}/merge/'.format(
                self.api_version, self.cl1.pk),
            ['http://testserver' + reverse(
                '{}:contacts:contactlist-detail'.format(self.api_version),
                kwargs={'pk': self.cl2.pk})],
            format='json')
        self.assertEqual(self.cl1.contacts.count(), 2)
        self.assertEqual(response.status_code, 200)
        self.assertFalse(ContactList.objects.filter(pk=self.cl2.pk).exists())
        self.assertFalse(Contact.objects.filter(pk=double_contact.pk).exists())


class ContactBulkAPITestCase(TestCase):
    def setUp(self):
        self.api_version = 'v1'
        self.user = UserFactory(groups=['managers'])
        self.client = APIClient()
        self.client.login(identifier=self.user.identifier, password='password')

    def test_two_valid_contacts(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {
                        'has_mustache': True,
                        'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {
                        'has_mustache': True,
                        'has_beard': False}
                },
            ], format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 2)

    def test_one_contact_not_valid(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {
                        'has_mustache': True,
                        'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {
                        'has_mustache': True}
                },
            ], format='json')

        self.assertEqual(400, response.status_code)
        self.assertIn(
            'test02@example.com', response.json()['validation_errors'].keys())
        self.assertEqual(response.json()['no_address'], '0')
        self.assertEqual(response.json()['duplicates'], [])
        self.assertEqual(Contact.objects.count(), 0)

    def test_duplicate_contacts(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
            ], format='json')

        self.assertEqual(400, response.status_code)
        self.assertIn('test02@example.com', response.json()['duplicates'])
        self.assertEqual(response.json()['no_address'], '0')
        self.assertEqual(response.json()['validation_errors'], {})
        self.assertEqual(Contact.objects.count(), 0)

    def test_invalid_and_duplicate_and_missing_email(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': 'test03@example.com',
                    'contact_list': url,
                    'properties': {'has_mustach': False}
                },
                {
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': '_no_addr',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                }
            ], format='json')

        self.assertEqual(400, response.status_code)
        self.assertIn('test03@example.com', response.json()[
            'validation_errors'])
        self.assertIn('test02@example.com', response.json()['duplicates'])
        self.assertEqual(response.json()['no_address'], '1')
        self.assertEqual(Contact.objects.count(), 0)

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': True}
                },
                {
                    'address': 'dadadaada??',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                }
            ], format='json')

        self.assertEqual(400, response.status_code)
        self.assertIn('dadadaada??', response.json()['validation_errors'])
        self.assertEqual(Contact.objects.count(), 0)

    def test_add_contacts_via_bulk(self):
        response = self.client.post(
            '/{}/contacts/lists/'.format(self.api_version),
            {
                'name': 'Test list 1',
                'contact_fields': [{
                    'name': 'has_beard',
                    'type': 'Boolean',
                    'required': True}],
                'source_type': 'api-test',
                'source_ref': '1'}, format='json')

        url = response.data['url']

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                },
                {
                    'address': 'test02@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                }
            ], format='json')

        self.assertEqual(201, response.status_code)
        self.assertEqual(Contact.objects.count(), 2)

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': True}
                },
                {
                    'address': 'dadadaada??',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                }
            ], format='json')

        self.assertEqual(400, response.status_code)
        self.assertEqual(Contact.objects.count(), 2)

        response = self.client.post(
            '/{}/contacts/'.format(self.api_version),
            [
                {
                    'address': 'test01@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': True}
                },
                {
                    'address': 'test03@example.com',
                    'contact_list': url,
                    'properties': {'has_beard': False}
                }
            ], format='json')

        self.assertEqual(400, response.status_code)
        self.assertIn('test01@example.com', response.json()[
            'validation_errors'])
        self.assertEqual(Contact.objects.count(), 2)



from django.conf.urls import url

from . import views


urlpatterns = [
    url(r'^report/(?P<identifier>[\w\-_]+)/$',
        views.abuse_report, name='abuse-report'),
    url(r'^thanks/(?P<identifier>[\w\-_]+)/$',
        views.abuse_report_thanks,
        name='abuse-report-thanks'),
]



from django.urls import reverse
from django.contrib import admin

from .models import Mail
from .models import Message
from .models import MessageAttachment
from .models import MailStatus


class MailStatusInline(admin.TabularInline):
    extra = 0
    max_num = 0
    can_delete = False
    model = MailStatus
    ordering = ['creation_date']
    readonly_fields = [
        'source_ip', 'destination_domain', 'status_code',
        'source_hostname', 'raw_msg', 'creation_date', 'status']


class MailInline(admin.TabularInline):
    extra = 0
    model = Mail
    readonly_fields = [
        'creation_date', 'first_status_date', 'latest_status_date',
        'delivery_duration', 'had_delay', 'identifier', 'curstatus']
    exclude = ['properties', 'source_type', 'source_ref']


class MessageAttachmentAdmin(admin.ModelAdmin):
    list_display = ['filename', 'message', 'size', 'b64size']
    list_filter = ['message']
    search_fields = ['message']
    raw_id_fields = ['message']
    readonly_fields = ['size', 'b64size']


class MessageAttachmentInline(admin.TabularInline):
    extra = 0
    model = MessageAttachment
    readonly_fields = ['filename', 'size', 'b64size', 'path']


class MessageAdmin(admin.ModelAdmin):
    list_display = [
        'name', 'author_url', 'status', 'creation_date',
        'send_date', 'completion_date', 'category_url']
    list_filter = ['status', 'creation_date', 'send_date']
    search_fields = ['name']
    inlines = [MessageAttachmentInline, MailInline]
    raw_id_fields = ['author']

    def author_url(self, instance):
        return '<a target="_blank" href="{}">{}</a>'.format(
            reverse('admin:users_munchuser_change', args=[instance.author.pk]),
            instance.author)
    author_url.allow_tags = True
    author_url.short_description = 'Author'

    def category_url(self, instance):
        if instance.category:
            return '<a target="_blank" href="{}">{}</a>'.format(
                reverse(
                    'admin:core_category_change', args=[instance.category.pk]),
                instance.category)
        else:
            return '-'
    category_url.allow_tags = True
    category_url.short_description = 'Category'


class MailStatusAdmin(admin.ModelAdmin):
    list_display = ['mail', 'status', 'creation_date']
    list_filter = ['status']


class MailAdmin(admin.ModelAdmin):
    list_display = [
        'recipient', 'creation_date', 'identifier',
        'last_status', 'message', 'had_delay']
    search_fields = ['recipient', 'identifier']
    inlines = [MailStatusInline]
    raw_id_fields = ['message']
    list_filter = ['had_delay']
    readonly_fields = [
        'creation_date', 'delivery_duration', 'last_status', 'had_delay',
        'first_status_date', 'latest_status_date', 'identifier']


admin.site.register(Mail, MailAdmin)
admin.site.register(Message, MessageAdmin)
admin.site.register(MessageAttachment, MessageAttachmentAdmin)
admin.site.register(MailStatus, MailStatusAdmin)



from django.conf import settings

from munch.core.mail.utils.munchers import ContentMuncherRunner
from munch.core.mail.utils.munchers import HeadersMuncherRunner

# Filters definitions
post_individual_html_generation = ContentMuncherRunner(
    settings.CAMPAIGNS, 'HTML_INDIVIDUAL_FILTERS')
post_template_html_generation = ContentMuncherRunner(
    settings.CAMPAIGNS, 'HTML_TEMPLATE_FILTERS')
post_individual_plaintext_generation = ContentMuncherRunner(
    settings.CAMPAIGNS, 'PLAINTEXT_INDIVIDUAL_FILTERS')
post_headers_generation = HeadersMuncherRunner(
    settings.CAMPAIGNS, 'HEADERS_FILTERS')



import os

import click


@click.command(
    add_help_option=False, context_settings=dict(ignore_unknown_options=True))
@click.argument('management_args', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def django(ctx, management_args):
    "Execute Django subcommands."
    if len(management_args) and management_args[0] == 'test':
        if '--settings' not in management_args:
            if os.environ['DJANGO_SETTINGS_MODULE'] == 'munch.settings':
                os.environ["DJANGO_SETTINGS_MODULE"] = "munch.settings.test"

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "munch.settings")

    from django.core.management import execute_from_command_line
    execute_from_command_line(argv=[ctx.command_path] + list(management_args))



from urllib.parse import urljoin

from django.urls import reverse
from django.conf import settings
from django.core.mail import EmailMessage
from django.template.loader import render_to_string

from munch.core.mail.utils import get_app_url


class PolicyBackend:
    def apply(self, item, policies_list):
        raise NotImplementedError()


class DoubleOptIn(PolicyBackend):
    def apply(self, item, policies_list):
        subscription_url = get_app_url(organization=item.get_owner())

        if 'BounceCheck' in policies_list:
            return_path = 'subscription-bounce+{uuid}@{fqdn}'.format(
                uuid=item.uuid, fqdn=settings.RETURNPATH_DOMAIN)
        else:
            return_path = settings.SERVICE_MSG_FROM_EMAIL
        confirmation_link = urljoin(
            subscription_url,
            reverse('confirmation', kwargs={'uuid': item.uuid}))
        message = EmailMessage(
            subject='Votre inscription',
            body=render_to_string(
                'contacts/double_opt_in_confirmation.email',
                {'contact': item, 'confirmation_link': confirmation_link}),
            to=(item.address, ),
            from_email='{} <{}>'.format(
                settings.SERVICE_MSG_FROM_NAME,
                settings.SERVICE_MSG_FROM_EMAIL),
            headers={
                'Auto-Submitted': 'auto-generated',
                'Return-Path': return_path})
        message.send()


class BounceCheck(PolicyBackend):
    def apply(self, item, policies_list):
        # if double opt-in is set up, we use the validation mail to check
        # bounce
        if 'DoubleOptIn' not in policies_list:
            return_path = 'subscription-bounce+{uuid}@{fqdn}'.format(
                uuid=item.uuid, fqdn=settings.RETURNPATH_DOMAIN)
            message = EmailMessage(
                subject='Votre inscription',
                body=render_to_string(
                    'contacts/bounce_check_confirmation.email',
                    {'contact': item}),
                to=(item.address, ),
                from_email='{} <{}>'.format(
                    settings.SERVICE_MSG_FROM_NAME,
                    settings.SERVICE_MSG_FROM_EMAIL),
                headers={
                    'Auto-Submitted': 'auto-generated',
                    'Return-Path': return_path})
            message.send()



from django.conf.urls import url
from django.conf.urls import include

from munch.core.utils.routers import APIRouter
from munch.urls import api_urlpatterns_v1

from . import views

router = APIRouter()
router.register('mails', views.MailViewSet, base_name='mail')
router.register('batches', views.MailBatchViewSet, base_name='batch')

api_urlpatterns_v1 += [
    url('transactional/', include(router.urls, namespace='transactional'))]



import urllib
from collections import OrderedDict

from rest_framework import relations
from rest_framework import serializers
from rest_framework.compat import unicode_to_repr
from rest_framework.fields import empty
from rest_framework.reverse import reverse
from django.core.exceptions import ValidationError as DjangoValidationError
from rest_framework.serializers import ValidationError as DRFValidationError


class HALLinksField(relations.HyperlinkedIdentityField):
    """ Tries to represent related links a HAL-compliant way.

    Offer special support for nested links.

      Ex: /foo/:n/info/ is a nested link of /foo/:n/

    See http://stateless.co/hal_specification.html
    """
    def __init__(self, endpoints={}, nested_endpoints=[], *args, **kwargs):
        """
        :param endpoints: dict of reversable urls to point to other API
                          resources, `link_name` will be used as key for the
                          URL in the resource.
        :type endpoints: dict {<link_name> {"lookup_key": str,
                               "lookup_field": str,
                               "view_name": str}
        :param nested_endpoints: list of url suffixes leading to nested
                                 operations on the resource
                                 (ex: ['preview', 'check'])
        """
        self.nested_endpoints = nested_endpoints
        self.endpoints = endpoints

        for d in endpoints.values():
            assert d.keys() == set(
                ['lookup_key', 'lookup_field', 'view_name']), (
                    "bad endpoint format keys set : {}".format(d.keys()))
        super().__init__(*args, **kwargs)

    def get_attribute(self, obj):
        return obj

    def to_representation(self, value):
        links = OrderedDict()
        prefix = super().to_representation(self.get_attribute(value))
        for i in self.nested_endpoints:
            # We consider if it contains a dot its a content-type indication,
            # so no trailing slash
            if '.' in i:
                suffix = ''
            else:
                suffix = '/'
            links[i] = {'href': urllib.parse.urljoin(prefix, i) + suffix}

        # Reverse all the URL for endpoints, using targeted object
        request = self.context.get('request', None)
        for url_name, url_lookup in self.endpoints.items():
            lookup_value = getattr(value, url_lookup['lookup_field'])
            links[url_name] = {
                'href':
                    reverse(
                        url_lookup['view_name'],
                        kwargs={url_lookup['lookup_key']: lookup_value},
                        request=request)}
        return links


class SizeLimitedListSerializer(serializers.ListSerializer):
    DEFAULT_MAX_ITEMS = 10000

    def run_list_size_validation(self, data):
        try:
            max_items = self.Meta.max_items()
        except AttributeError:
            max_items = self.DEFAULT_MAX_ITEMS

        items_count = len(data)

        if items_count > max_items:
            raise serializers.ValidationError(detail={'non_field_errors': [
                'Serializer limited to {} items at once, got {}'.format(
                    max_items, items_count)]})

    def run_validation(self, data=empty):
        """
        Natural place for this validation to occur would be in self.validate()
        but it's not a safe option as we would validate *all* items *before*
        rejecting.

        Overloading run_validation allows to raise ValidationError earlier.
        """
        self.run_list_size_validation(data)
        return super().run_validation(data)

    # Those two are bugfix for DRF bulk:
    # https://github.com/miki725/django-rest-framework-bulk/issues/33
    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration


class CurrentOrganizationDefault(object):
    """ Class inspired by rest_framework.serializers.CurrentUserDefault
    """
    def set_context(self, serializer_field):
        request = serializer_field.context['request']
        if request.user.is_authenticated():
            self.organization = request.user.organization
        else:
            self.organization = None

    def __call__(self):
        return self.organization

    def __repr__(self):
        return unicode_to_repr('%s()' % self.__class__.__name__)


class ForwardValidationErrorsMixin:
    """ Allows to forward django validation errors as DRF validation errors,
    on save
    """
    def create(self, validated_data, *args, **kwargs):
        try:
            return super().create(validated_data, *args, **kwargs)
        except DjangoValidationError as e:
            raise DRFValidationError(e.error_list)

    def update(self, instance, validated_data, *args, **kwargs):
        try:
            return super().update(instance, validated_data, *args, **kwargs)
        except DjangoValidationError as e:
            raise DRFValidationError(e.error_list)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-06 10:05
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.hstore
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import munch.apps.campaigns.fields
import munch.apps.campaigns.models
import munch.core.mail.utils
import munch.core.utils.models
from django.contrib.postgres.operations import HStoreExtension


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0002_permissions'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        HStoreExtension(),
        migrations.CreateModel(
            name='Mail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient', models.EmailField(max_length=254, verbose_name='recipient')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now, verbose_name='creation date')),
                ('first_status_date', models.DateTimeField(blank=True, null=True, verbose_name='first status date')),
                ('latest_status_date', models.DateTimeField(blank=True, null=True, verbose_name='latest status date')),
                ('delivery_duration', models.DurationField(null=True, verbose_name='delivery duration')),
                ('had_delay', models.BooleanField(default=False, verbose_name='had delay')),
                ('identifier', models.CharField(db_index=True, default=munch.apps.campaigns.models.get_base_mail_identifier, max_length=35, unique=True, verbose_name='identifier')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='properties')),
                ('source_type', models.CharField(blank=True, max_length=100, validators=[django.core.validators.RegexValidator(message='source should be composed of lowercase letters and dashes', regex='^[a-z-]*$')], verbose_name='source type')),
                ('source_ref', models.CharField(blank=True, max_length=200, verbose_name='source reference')),
                ('curstatus', munch.apps.campaigns.fields.FSMAutoField(choices=[('unknown', 'unknown'), ('ignored', 'ignored'), ('deleted', 'deleted'), ('queued', 'queued'), ('sending', 'sending'), ('delayed', 'delayed (retrying)'), ('delivered', 'delivered to remote MTA'), ('dropped', 'not delivered (soft-bounce)'), ('bounced', 'not delivered (hard-bounce)')], default='unknown', max_length=50, verbose_name='current status')),
            ],
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='MailStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('unknown', 'unknown'), ('ignored', 'ignored'), ('deleted', 'deleted'), ('queued', 'queued'), ('sending', 'sending'), ('delayed', 'delayed (retrying)'), ('delivered', 'delivered to remote MTA'), ('dropped', 'not delivered (soft-bounce)'), ('bounced', 'not delivered (hard-bounce)')], default='queued', max_length=15, verbose_name='status')),
                ('creation_date', models.DateTimeField(verbose_name='creation_date')),
                ('source_ip', models.GenericIPAddressField(blank=True, null=True, verbose_name='source ip')),
                ('source_hostname', models.CharField(blank=True, max_length=100, null=True, verbose_name='source hostname')),
                ('destination_domain', models.CharField(max_length=150, verbose_name='destination domain')),
                ('status_code', models.CharField(blank=True, help_text='X.XXX.XXX, as defined in RFC3463', max_length=25, validators=[django.core.validators.RegexValidator(message='status code should follow rfc3463', regex='([234](\\.\\d{1,3}){2})')], verbose_name='SMTP status code')),
                ('raw_msg', models.TextField(verbose_name='raw message')),
                ('mail', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='statuses', to='campaigns.Mail', verbose_name='mail')),
            ],
            options={
                'verbose_name': 'email status',
                'verbose_name_plural': 'email statuses',
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
                'abstract': False,
                'get_latest_by': 'creation_date',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(db_index=True, default=munch.core.mail.utils.mk_base64_uuid, editable=False, max_length=35, unique=True, verbose_name='identifier')),
                ('name', models.CharField(max_length=60, verbose_name='name')),
                ('sender_email', models.EmailField(max_length=254, verbose_name='sender address')),
                ('sender_name', models.CharField(max_length=50, verbose_name='sender name')),
                ('subject', models.CharField(blank=True, max_length=130, verbose_name='subject')),
                ('html', models.TextField(blank=True)),
                ('status', munch.apps.campaigns.fields.FSMAutoField(choices=[('new', 'Created'), ('message_ok', 'Message approved'), ('message_issues', 'Message contains issues'), ('sending', 'Sending'), ('sent', 'Sent')], default='new', max_length=50, verbose_name='status')),
                ('creation_date', models.DateTimeField(auto_now_add=True, help_text='automatic', verbose_name='creation date')),
                ('send_date', models.DateTimeField(blank=True, null=True, verbose_name='send date')),
                ('completion_date', models.DateTimeField(blank=True, null=True, verbose_name='complete sent date')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='properties')),
                ('external_optout', models.BooleanField(default=False, help_text='If enabled, this message, instead of optout form, recipients will be invited to contact you.', verbose_name='external optout')),
                ('detach_images', models.BooleanField(default=False, help_text='Host linked images on our servers instead of original links.', verbose_name='host images')),
                ('track_open', models.BooleanField(default=False, help_text='Detect who opens message by adding hidden image.', verbose_name='track opens')),
                ('track_clicks', models.BooleanField(default=False, help_text='Detect which links were clicked by replacing original links by redirections', verbose_name='track clicks')),
                ('spam_score', models.FloatField(blank=True, null=True, verbose_name='spam score')),
                ('spam_details', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True, verbose_name='Triggered anti-spam rules')),
                ('is_spam', models.BooleanField(default=False, verbose_name='Detected as spam')),
                ('spam_check_error', models.TextField(blank=True, null=True, verbose_name='Spam check error')),
                ('msg_issue', models.CharField(blank=True, max_length=300, null=True, verbose_name='message issue')),
                ('msg_links', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='message links')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='core.Category', verbose_name='category')),
            ],
            options={
                'verbose_name': 'message',
                'verbose_name_plural': 'messages',
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations', 'previewsend_organizations', 'previewsend_mine', 'previewsend'),
                'abstract': False,
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='MessageAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', models.FileField(upload_to=munch.apps.campaigns.models.get_attachment_storage_path, verbose_name='attached file')),
                ('original_name', models.CharField(max_length=100, verbose_name='original name')),
                ('b64size', models.PositiveIntegerField(default=0, verbose_name='base 64 size')),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='campaigns.Message', verbose_name='message')),
            ],
            options={
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
                'abstract': False,
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='PreviewMail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('recipient', models.EmailField(max_length=254, verbose_name='recipient')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now, verbose_name='creation date')),
                ('first_status_date', models.DateTimeField(blank=True, null=True, verbose_name='first status date')),
                ('latest_status_date', models.DateTimeField(blank=True, null=True, verbose_name='latest status date')),
                ('delivery_duration', models.DurationField(null=True, verbose_name='delivery duration')),
                ('had_delay', models.BooleanField(default=False, verbose_name='had delay')),
                ('identifier', models.CharField(db_index=True, default=munch.apps.campaigns.models.get_base_mail_identifier, max_length=35, unique=True, verbose_name='identifier')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(blank=True, null=True, verbose_name='properties')),
                ('source_type', models.CharField(blank=True, max_length=100, validators=[django.core.validators.RegexValidator(message='source should be composed of lowercase letters and dashes', regex='^[a-z-]*$')], verbose_name='source type')),
                ('source_ref', models.CharField(blank=True, max_length=200, verbose_name='source reference')),
                ('curstatus', munch.apps.campaigns.fields.FSMAutoField(choices=[('unknown', 'unknown'), ('ignored', 'ignored'), ('deleted', 'deleted'), ('queued', 'queued'), ('sending', 'sending'), ('delayed', 'delayed (retrying)'), ('delivered', 'delivered to remote MTA'), ('dropped', 'not delivered (soft-bounce)'), ('bounced', 'not delivered (hard-bounce)')], default='unknown', max_length=50, verbose_name='current status')),
                ('message', munch.apps.campaigns.models.TolerantForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='previewmails', to='campaigns.Message', verbose_name='message')),
            ],
            options={
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
                'abstract': False,
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.AddField(
            model_name='mail',
            name='message',
            field=munch.apps.campaigns.models.TolerantForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mails', to='campaigns.Message', verbose_name='message'),
        ),
        migrations.AlterUniqueTogether(
            name='mail',
            unique_together=set([('recipient', 'message')]),
        ),
    ]



import urllib

from django.urls import get_resolver
from django.urls import NoReverseMatch


def resolve_url(request, path=None):
    """Helper function that reports information on the request's url.

    Taken from http://code.google.com/p/greatlemers-django-tools \
                /source/browse/trunk/gdt_nav/models.py#158
    Apache License 2.0

    This utility function takes a request and analyses its url to generate the
    url_name and keyword arguments that can be used to generate the url via
    the reverse function or url tag.
    The url resolver doesn't return the name of the url that produces the
    given url so some hunting around has to be done to determine what exactly
    it should be.

    Keyword arguments:
    request -- The request object for the view that wants to generate some
    menus.
    path    -- The relative path (default: gets it from the request.path_info)

    Returns:
    A tuple of (url, url_name, url_kwargs)
    url -- The absolute representation of the requested url
    url_name -- The 'reversable' name of the requested url
    url_kwargs -- The keyword arguments that would be needed in order to
    'reverse' the url.
    """
    # Start by fetching the path from the request and using it to build
    # the full url.
    if not path:
        path = request.path_info

    url = request.build_absolute_uri(path)
    # make sure path is only the local path
    path = urllib.parse.urlparse(path).path

    # The url resolver which will generate some of the url info.
    # Get urlconf from request object if available.
    urlconf = getattr(request, "urlconf", None)
    resolver = get_resolver(urlconf)

    # Pull out the view function, and the url arguments and keywords.
    view_func, url_args, url_kwargs = resolver.resolve(path)

    # Fetch a list of all the signatures of the items that can be reversed
    # to produce the view function.
    sigs = resolver.reverse_dict.getlist(view_func)

    url_name = None
    # Loop through all the items in the reverse dictionary.
    for key, value in resolver.reverse_dict.items():
        # Check if the value of the mapping is one of our matching signatures
        # and that the key is a string.
        if value in sigs and type(key) == str:
            try:
                # See if we have the right parameters to use this reversal and
                # that it produces the correct url.
                if resolver.reverse(key, *url_args, **url_kwargs) == path[1:]:
                    # No exceptions were thrown so we have the right parameters
                    # and the path matched therefore we've found the url name
                    # we were seeking - which of course means we can
                    # stop looking.
                    url_name = key
                    break
            except NoReverseMatch:
                # The parameters were wrong - ah well, maybe the next one will
                # succeed.
                pass
    return url, url_name, url_kwargs



import factory.django
from faker import Factory as FakerFactory

from munch.apps.users.tests.factories import OrganizationFactory

from ..models import SendingDomain

faker = FakerFactory.create()


class SendingDomainFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = SendingDomain

    name = factory.LazyAttribute(lambda x: faker.domain_name())
    organization = factory.SubFactory(OrganizationFactory)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-23 17:53
from __future__ import unicode_literals

from django.db import migrations, models
import munch.apps.domains.fields
import munch.core.utils.validators


class Migration(migrations.Migration):

    dependencies = [
        ('domains', '0002_permissions'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sendingdomain',
            name='app_domain',
            field=models.CharField(blank=True, default='', help_text='Domain to use for email links generation', max_length=200, validators=[munch.core.utils.validators.DomainNameValidator()], verbose_name='Custom application domain'),
        ),
        migrations.AlterField(
            model_name='sendingdomain',
            name='app_domain_status',
            field=munch.apps.domains.fields.DomainCheckField(choices=[('ok', 'Configured'), ('ko', 'Not configured'), ('bad', 'Badly configured'), ('pending', 'Checking'), ('unknown', 'Unknown')], default='unknown', max_length=10, verbose_name='Custom application domain status'),
        ),
        migrations.AlterField(
            model_name='sendingdomain',
            name='app_domain_status_date',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Custom application domain status last change'),
        ),
    ]



import shutil

from django.test import TestCase
from django.test.utils import override_settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage

from munch.apps.users.tests.factories import UserFactory
from munch.apps.campaigns.models import MessageAttachment
from munch.apps.campaigns.tests.factories import MessageFactory


class MessageAttachmentTest(TestCase):
    @override_settings(BYPASS_DNS_CHECKS=True)
    def setUp(self):
        self.user = UserFactory()
        self.message = MessageFactory(author=self.user)

    def tearDown(self):
        shutil.rmtree(default_storage.location)

    def test_delete_message_with_shared_attachment(self):
        default_storage.save('foo.txt', ContentFile('testfoo'))

        message_attachment = MessageAttachment.objects.create(
            message=self.message, file='foo.txt')

        another_message = MessageFactory(author=self.user)

        another_message_attachment = MessageAttachment.objects.create(
            message=another_message, file='foo.txt')

        self.assertTrue(default_storage.exists(message_attachment.file.path))
        self.assertTrue(default_storage.exists(
            another_message_attachment.file.path))

        self.message.delete()

        self.assertTrue(default_storage.exists(message_attachment.file.path))
        self.assertTrue(default_storage.exists(
            another_message_attachment.file.path))

        another_message.delete()

        self.assertFalse(default_storage.exists(message_attachment.file.path))
        self.assertFalse(default_storage.exists(
            another_message_attachment.file.path))



from rest_framework.exceptions import APIException


class InvalidSubmitedData(ValueError, APIException):
    status_code = 500
    default_detail = 'Some data is invalid'

    @property
    def detail(self):
        return self.__str__()


class WrongHTML(InvalidSubmitedData):
    def __init__(self, parent_e):
        self._e = parent_e
        super().__init__()

    def __str__(self):
        return 'Le HTML soumis est invalide : "{}"'.format(str(self._e))



import os
import sys
import logging
from functools import wraps

import celery
from celery.signals import celeryd_after_setup
from kombu import Queue

from django.conf import settings

from munch.core.utils import get_worker_type

log = logging.getLogger('munch')
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'munch.settings')


def catch_exception(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except Exception as err:
            sys.stderr.write(str(err))
            raise
    return wrapper


class Celery(celery.Celery):
    def on_configure(self):
        if hasattr(settings, 'RAVEN_CONFIG'):
            from raven import Client
            from raven.contrib.celery import register_signal
            from raven.contrib.celery import register_logger_signal
            client = Client(settings.RAVEN_CONFIG.get('dsn'))
            register_logger_signal(client)
            register_signal(client)


class CeleryRouteMap(object):
    def __init__(self, app):
        self.app = app

        self.exchange = settings.CELERY_DEFAULT_EXCHANGE
        self.exchange_type = settings.CELERY_DEFAULT_EXCHANGE_TYPE

        self.queues = {
            'default': {
                'name': settings.CELERY_DEFAULT_QUEUE,
                'routing_key': settings.CELERY_DEFAULT_ROUTING_KEY
            }
        }
        self.routes = {}

    def add_queue(self, worker_type, queue):
        self.app.amqp.queues.add(
            Queue(
                queue, routing_key='{}.#'.format(queue),
                queue_arguments={'x-max-priority': 100}
            )
        )
        self.queues.update({
            worker_type: {'name': queue, 'routing_key': queue}})
        log.debug('Added queue {} for {} workers'.format(
            queue, worker_type.upper()))

    def register_route(self, task, worker_type, munch_app):
        if worker_type not in self.queues:
            raise ValueError(
                'Can not register celery route. '
                'No queue defined for worker_type {}'.format(
                    worker_type.upper()))
        self.routes.update({task: {'worker': worker_type, 'key': munch_app}})
        log.debug(
            'Registered route for {} on {} workers'.format(
                task, worker_type.upper()))

    def import_tasks_map(self, tasks_map, munch_app):
        for worker, tasks in tasks_map.items():
            for task in tasks:
                self.register_route(task, worker, munch_app)

    def lookup_route(self, task):
        if task in self.routes:
            worker = self.routes.get(task)['worker']
            key = self.routes.get(task)['key']
            queue = self.queues.get(worker)
            return {
                'queue': queue['name'],
                'exchange': self.exchange,
                'exchange_type': self.exchange_type,
                'routing_key': '{}.{}'.format(queue['routing_key'], key)
            }
        return None

    def register_to_queue(self, queue):
        self.app.amqp.queues.select_add(
            queue, routing_key='{}.#'.format(queue),
            queue_arguments={'x-max-priority': 100})

    def register_as_worker(self, worker_type):
        if worker_type not in self.queues:
            raise ValueError(
                'Can not register as worker {}. '
                'No queue defined for this worker_type'.format(
                    worker_type.upper()))
        self.register_to_queue(self.queues[worker_type]['name'])

    def get_queue_for(self, worker_type):
        return self.queues.get(worker_type, 'default')['name']

    def get_workers_map(self):
        workers_map = {}
        for k, v in self.routes.items():
            workers_map.setdefault(v['worker'], []).append(k)
        return workers_map


class CeleryRouter(object):
    def route_for_task(self, task, args=None, kwargs=None):
        return munch_tasks_router.lookup_route(task)


# Celery App initialization
app = Celery('munch', broker=settings.BROKER_URL)
app.config_from_object('django.conf:settings')

munch_tasks_router = CeleryRouteMap(app)


# Queues, Tasks and worker registration methods for munch.core
def add_queues():
    munch_tasks_router.add_queue('core', 'munch.core')
    munch_tasks_router.add_queue('status', 'munch.status')
    munch_tasks_router.add_queue('gc', 'munch.gc')


def register_tasks():
    tasks_map = {
        'gc': ['munch.core.mail.tasks.purge_raw_mail']
    }
    munch_tasks_router.import_tasks_map(tasks_map, 'munch')


@celeryd_after_setup.connect
@catch_exception
def configure_worker(instance, **kwargs):
    if get_worker_type() in ['gc', 'all']:
        from .mail.tasks import purge_raw_mail  # noqa
        sys.stdout.write(
            '[core-app] Registering worker as GARBAGE COLLECTOR...')
        munch_tasks_router.register_as_worker('gc')



from django.conf import settings

from munch.core.mail.utils.munchers import ContentMuncherRunner

# Munchers on the content that is displayed in HTML web version
# (served via HTTP)
post_web_html_individual = ContentMuncherRunner(
    settings.CAMPAIGNS, 'WEB_HTML_INDIVIDUAL_FILTERS')



import functools

import lxml.etree
from lxml.cssselect import CSSSelector
from django.urls import reverse

from munch.core.utils.regexp import mkd_footnote_url_re

from .utils import WebKey

body_selector = CSSSelector('body')
links_selector = CSSSelector('a')


def mk_tracking_url(app_url, mail_identifier):
    """ Builds the tracking image at that url

    :type mail: campaigns.Mail
    :returns: the absolute url of the tracking image
    """
    return '{}/{}'.format(
        app_url, reverse(
            'tracking-open',
            kwargs={'identifier': mail_identifier}).strip('/'))


def add_tracking_image(
        html, app_url=None, track_open=False, mail_identifier=None, **kwargs):
    """ Content filter to add tracking Pixel

    Appends a <img> tag to the HTML, pointing to a tracking with id in
    querystring
    """
    if track_open and app_url and mail_identifier:
        # use the root tree to preserve doctype
        doc = lxml.etree.HTML(html).getroottree()
        img = lxml.etree.fromstring(
            '<img src="{}" alt="" height="1" width="1" border="0" />'.format(
                mk_tracking_url(app_url, mail_identifier)))

        body = body_selector(doc)

        if len(body) > 0:
            body[0].append(img)
        else:
            doc.append(img)

        with_image = lxml.html.tostring(doc).decode()
        return with_image
    return html


class LinksRewriter:
    @staticmethod
    def rewrite(mail_identifier, app_url, links_map, url):
        """ Makes a redirection url used for email links
        """
        # invert the message links_map so we can look by URL
        if links_map:
            links_map = {v: k for k, v in links_map.items()}
        if url in links_map:
            return '{}/{}'.format(
                app_url,
                reverse('tracking-redirect',
                        kwargs={
                            'identifier': mail_identifier,
                            'link_identifier': links_map[url]}).strip('/'))
        return url

    @staticmethod
    def should_rewrite(link, unsubscribe_url):
        """ Just a place to store a blacklist of links not to be rewritten

        FIXME: can be removed I think
        """
        if link == unsubscribe_url:
            return False
        return True


class HTMLLinksRewriter(LinksRewriter):
    @classmethod
    def _rewrite_html_links(
            cls, html, track_clicks, mail_identifier,
            unsubscribe_url, app_url, links_map, rewrite_func):
        """ Generic Content filter to rewrite links

        :param rewrite_func: f(Mail, original_url) -> rewritten_url
        """
        if track_clicks and \
                not mail_identifier.startswith('test'):
            # use the root tree to preserve doctype
            doc = lxml.etree.HTML(html).getroottree()

            for link in links_selector(doc):
                original_url = link.get('href') or ''
                original_url = original_url.strip()

                if (
                        original_url.startswith('http') and cls.should_rewrite(
                            original_url, unsubscribe_url)):
                    link.set('href', rewrite_func(
                        mail_identifier, app_url,
                        links_map, original_url))

            return lxml.html.tostring(doc).decode()
        return html


class HTMLEMailLinksRewriter(HTMLLinksRewriter):
    def __call__(
            self, html, app_url=None, track_clicks=False,
            mail_identifier=None, unsubscribe_url=None,
            links_map={}, **kwargs):
        """ Content filter to add tracked links for mails """
        return self._rewrite_html_links(
            html, track_clicks, mail_identifier,
            unsubscribe_url, app_url, links_map, self.rewrite)


rewrite_html_links = HTMLEMailLinksRewriter()


class PlaintextLinksRewriter(LinksRewriter):
    def __call__(
            self, plaintext, app_url=None, unsubscribe_url=None,
            mail_identifier=None, links_map={}, **kwargs):
        rewriter = functools.partial(
            self._rewrite_link,
            app_url=app_url,
            links_map=links_map,
            unsubscribe_url=unsubscribe_url,
            mail_identifier=mail_identifier)
        return mkd_footnote_url_re.sub(rewriter, plaintext)

    @classmethod
    def _rewrite_link(
            cls, m, app_url, links_map, unsubscribe_url, mail_identifier):
        url = m.group('url')
        footnote_mark = m.group('fnmark')
        if cls.should_rewrite(url, unsubscribe_url) and \
                not mail_identifier.startswith('test'):
            return '{} {}'.format(footnote_mark, cls.rewrite(
                mail_identifier, app_url, links_map, url))
        return m.group(0)  # as-is


rewrite_plaintext_links = PlaintextLinksRewriter()


class WebVersionLinksRewriter(HTMLLinksRewriter):
    def __call__(
            self, html, app_url=None, track_clicks=False,
            unsubscribe_url=None, mail_identifier=None,
            links_map={}, **kwargs):
        return self._rewrite_html_links(
            html, track_clicks, mail_identifier, unsubscribe_url,
            app_url, links_map, self.rewrite)

    @staticmethod
    def rewrite(mail_identifier, app_url, links_map, url):
        """Makes a redirection url used for links in web version

        It's based on web key.
        """
        webkey = WebKey.from_instance(mail_identifier)
        # invert the message links_map so we can look by URL
        links_map = {v: k for k, v in links_map.items()}

        if url in links_map:
            return '{}/{}'.format(
                app_url,
                reverse('web-tracking-redirect',
                        kwargs={
                            'web_key': webkey.token,
                            'link_identifier': links_map[url]}).strip('/'))
        return url


web_rewrite_html_links = WebVersionLinksRewriter()



# flake8: noqa
import os
import sys
import logging
from ssl import SSLContext
from ssl import PROTOCOL_TLSv1_2  # TODO: Switch to PROTOCOL_TLS with Py3.5
from multiprocessing import cpu_count

import click

log = logging.getLogger('munch.apps.transactional')


@click.group()
def run():
    "Run a service."


@run.command()
def smtp():
    "Run smtp smarthost for transactional service."
    from gevent import monkey
    monkey.patch_all()

    import os
    import sys

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "munch.settings")

    from psycogreen.gevent import patch_psycopg
    patch_psycopg()

    sys.path.append(os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..')))

    import django
    django.setup()

    import time
    import logging

    from django.conf import settings

    from gevent import sleep
    from gevent.pool import Pool
    from slimta.system import drop_privileges

    from munch.core.utils import monkey_patch_slimta_exception
    from munch.apps.transactional.queue import queue
    from munch.apps.transactional.edge import (
        TransactionalSmtpEdge, ProxyProtocolTransactionalSmtpEdge)
    from munch.apps.transactional.edge import EdgeValidators

    monkey_patch_slimta_exception()

    tls_settings = settings.TRANSACTIONAL.get('SMTP_SMARTHOST_TLS')

    if tls_settings is not None:
        if not isinstance(tls_settings, dict):
            raise Exception(
                ('Setting SMTP_SMARTHOST_TLS should be a dict with '
                 '"keyfile" and "certfile" keys'))

        for i in ('keyfile', 'certfile'):
            if i not in tls_settings:
                raise Exception(
                    '{} is required if you plan to offer STARTTLS support')
            else:
                path = tls_settings[i]
                if not os.access(path, os.R_OK):
                    raise Exception(
                        '{} is not readable or inexistant.'.format(path))
        ssl_context = SSLContext(PROTOCOL_TLSv1_2)
        ssl_context.load_cert_chain(
            tls_settings.get('certfile'), keyfile=tls_settings.get('keyfile'))
    else:
        ssl_context = None

    pool = Pool(settings.TRANSACTIONAL.get('EDGE_MAX_CONN', 200))

    edge_class = TransactionalSmtpEdge
    if settings.TRANSACTIONAL.get('PROXYPROTO_ENABLED', False):
        edge_class = ProxyProtocolTransactionalSmtpEdge

    edge = edge_class(
        (
            settings.TRANSACTIONAL.get('SMTP_BIND_HOST'),
            settings.TRANSACTIONAL.get('SMTP_BIND_PORT')),
        queue,
        data_timeout=settings.TRANSACTIONAL.get(
            'EDGE_TIMEOUTS', {}).get('data_timeout'),
        command_timeout=settings.TRANSACTIONAL.get(
            'EDGE_TIMEOUTS', {}).get('command_timeout'),
        pool=pool,
        hostname=settings.TRANSACTIONAL.get('EDGE_EHLO_AS', None),
        validator_class=EdgeValidators,
        context=ssl_context,
        auth=True,
    )

    log.info('Listening on {}:{}'.format(
        settings.TRANSACTIONAL.get('SMTP_BIND_HOST'),
        settings.TRANSACTIONAL.get('SMTP_BIND_PORT')))
    edge.start()

    if settings.TRANSACTIONAL.get('DROP_PRIVILEGES_USER') is not None:
        # ??? Really needed? See python-slimta/examples/slimta-mail.py...
        sleep(0.1)

        # If this command is run with root user (to be allowed to
        # open reserved ports like 25), we should then "switch" to a
        # normal user for security.
        drop_privileges(
            settings.TRANSACTIONAL.get('DROP_PRIVILEGES_USER'),
            settings.TRANSACTIONAL.get('DROP_PRIVILEGES_GROUP'))
        log.info('Dropping privileges to {}:{}'.format(
            settings.TRANSACTIONAL.get('DROP_PRIVILEGES_USER'),
            settings.TRANSACTIONAL.get('DROP_PRIVILEGES_GROUP')))

    try:
        edge.get()
    except KeyboardInterrupt:
        try:
            stop_timeout = settings.TRANSACTIONAL.get('SMTP_STOP_TIMEOUT', 5)
            edge.server.close()
            ts = time.time()
            log.info('Stop accepting connections.')
            if stop_timeout > 0:
                log.info(
                    'Waiting {} seconds before timing out '
                    'current connections...'.format(
                        stop_timeout))
                while time.time() - ts <= stop_timeout:
                    sleep(1.0)
            log.info('Edge stopped...')
            edge.server.stop(timeout=1)
        except KeyboardInterrupt:
            log.info('Forcing shutdown...')
            edge.server.stop(timeout=1)


@run.command()
@click.option('--pool', '-P', help=(
    'Pool implementation: prefork (default), '
    'eventlet, gevent, solo or threads.'), default='prefork')
@click.option('--hostname', '-n', help=(
    'Set custom hostname, e.g. \'w1.%h\'. Expands: %h'
    '(hostname), %n (name) and %d, (domain).'))
@click.option('--concurrency', '-c', default=cpu_count(), help=(
    'Number of child processes or threads to spawn. The '
    'default is the number of CPUs available on your system.'))
@click.option('--quiet', '-q', is_flag=True, default=False)
@click.option('--no-color', is_flag=True, default=False)
@click.option(
    '--autoreload', is_flag=True,
    default=False, help='Enable autoreloading.')
@click.option(
    '--worker-type', default='all',
    type=click.Choice(['all', 'core', 'status', 'router', 'mx', 'gc']),
    help=('Define which kind of task worker will consume.'),
    show_default=True)
@click.option(
    '--loglevel', '-l', help='Logging level',
    type=click.Choice(
        ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', 'FATAL']))
def worker(**options):
    "Run background worker instance."
    from django.conf import settings
    if hasattr(settings, 'CELERY_ALWAYS_EAGER') and \
            settings.CELERY_ALWAYS_EAGER:
        raise click.ClickException(
            'Disable CELERY_ALWAYS_EAGER in your '
            'settings file to spawn workers.')

    from munch.core.celery import app
    worker_type = options.pop('worker_type')
    os.environ['WORKER_TYPE'] = worker_type.upper()
    pool_cls = options.pop('pool')
    worker = app.Worker(
        pool_cls=pool_cls, queues=settings.CELERY_DEFAULT_QUEUE, **options)
    worker.start()
    try:
        sys.exit(worker.exitcode)
    except AttributeError:
        # `worker.exitcode` was added in a newer version of Celery:
        # https://github.com/celery/celery/commit/dc28e8a5
        # so this is an attempt to be forwards compatible
        pass


@run.command()
@click.option('--quiet', '-q', is_flag=True, default=False)
@click.option('--no-color', is_flag=True, default=False)
@click.option(
    '--autoreload', is_flag=True, default=False, help='Enable autoreloading.')
@click.option(
    '--loglevel', '-l', help='Logging level',
    type=click.Choice(
        ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', 'FATAL']))
def cron(**options):
    "Run periodic task dispatcher."
    from django.conf import settings
    if hasattr(settings, 'CELERY_ALWAYS_EAGER') and \
            settings.CELERY_ALWAYS_EAGER:
        raise click.ClickException(
            'Disable CELERY_ALWAYS_EAGER in your '
            'settings file to spawn workers.')

    from munch.core.celery import app
    app.Beat(**options).run()


@run.command()
def backmuncher():
    "Run smtp that handle feedback loops, unsuscribes and more."

    import django
    django.setup()

    from django.conf import settings
    from slimta.edge.smtp import SmtpEdge
    from slimta.system import drop_privileges
    from slimta.util.proxyproto import ProxyProtocol
    from slimta.relay.blackhole import BlackholeRelay

    from munch.core.mail.backmuncher import Queue

    edge_class = SmtpEdge
    if settings.BACKMUNCHER.get('PROXYPROTO_ENABLED', False):
        class ProxyProtocolSmtpEdge(ProxyProtocol, SmtpEdge):
            pass
        edge_class = ProxyProtocolSmtpEdge

    edge = edge_class(
        (
            settings.BACKMUNCHER.get('SMTP_BIND_HOST'),
            settings.BACKMUNCHER.get('SMTP_BIND_PORT')),
        Queue(BlackholeRelay()),
        hostname=settings.BACKMUNCHER.get('EDGE_EHLO_AS', None),
        tls=False,
    )

    log.info('Listening on {}:{}'.format(
        settings.BACKMUNCHER.get('SMTP_BIND_HOST'),
        settings.BACKMUNCHER.get('SMTP_BIND_PORT')))
    edge.start()

    if settings.BACKMUNCHER.get('DROP_PRIVILEGES_USER') is not None:
        # If this command is run with root user (to be allowed to
        # open reserved ports like 25), we should then "switch" to a
        # normal user for security.
        drop_privileges(
            settings.BACKMUNCHER.get('DROP_PRIVILEGES_USER'),
            settings.BACKMUNCHER.get('DROP_PRIVILEGES_GROUP'))
        log.info('Dropping privileges to {}:{}'.format(
            settings.BACKMUNCHER.get('DROP_PRIVILEGES_USER'),
            settings.BACKMUNCHER.get('DROP_PRIVILEGES_GROUP')))

    try:
        edge.get()
    except KeyboardInterrupt:
        try:
            edge.server.close()
            log.info('Stop accepting connections.')
            log.info('Edge stopped...')
            edge.server.stop(timeout=1)
        except KeyboardInterrupt:
            log.info('Forcing shutdown...')
            edge.server.stop(timeout=1)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-23 13:16
from __future__ import unicode_literals

from django.db import migrations


def comments_default(apps, schema_editor):
    AbuseNotification = apps.get_model('abuse', 'AbuseNotification')
    for report in AbuseNotification.objects.all():
        if not report.comments:
            report.comments = 'null'
            report.save()


class Migration(migrations.Migration):

    dependencies = [
        ('abuse', '0002_auto_20160923_1516'),
    ]

    operations = [
        migrations.RunPython(comments_default)
    ]



import factory.django
from faker import Factory as FakerFactory

from ..models import Category

faker = FakerFactory.create()


class CategoryFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Category

    name = factory.LazyAttribute(lambda x: faker.word())



from datetime import datetime
from unittest.mock import patch
from unittest.mock import MagicMock

import dns
import pytz
from django.test import TestCase
from django.conf import settings
from django.core.exceptions import ValidationError
from django.test.utils import override_settings
from libfaketime import fake_time

from munch.apps.users.tests.factories import UserFactory
from munch.apps.campaigns.tests.factories import MessageFactory

from ..utils import SPFValidator
from ..models import SendingDomain
from ..fields import DomainCheckField
from ..tasks import run_domains_validation
from ..tasks import validate_sending_domain_field

from .factories import SendingDomainFactory


def DNSResultMock(v):
    """ Mocks DNS, single valued response

    DNSResultMock(42)
    DNSResultMock[1].to_text()
    """
    m = MagicMock()
    sub_mock = MagicMock()
    sub_mock.to_text.return_value = v
    m.__getitem__.return_value = sub_mock
    m.__iter__.return_value = [sub_mock]
    return m


def DNSEmptyEntryMock():
    m = MagicMock()
    m.__getitem__.side_effect = IndexError()
    return m


def mock_resolver(queries):
    mock_queries = queries

    def mock_resolve(name, field):
        try:
            res = mock_queries[name][field]
        except KeyError:
            raise dns.exception.DNSException
        else:
            if getattr(res, 'side_effect', None):
                # Raise exception if needed
                return res()
            else:
                return res
    m = MagicMock()
    m.side_effect = mock_resolve
    return m

mock_queries = {
    'test._domainkey.sandbox.munch.example.com': {
        'TXT': DNSResultMock(settings.DOMAINS['DKIM_KEY_CONTENT'])},
    'sandbox.munch.example.com': {
        'TXT': DNSResultMock('"v=spf1 include:spf.munch.example.com -all"')},  # noqa
    'test._domainkey.bad.sandbox.munch.example.com': {
        'TXT': DNSResultMock(settings.DOMAINS[
            'DKIM_KEY_CONTENT'].replace('1', '2'))},
    'bad.sandbox.munch.example.com': {
        'TXT': DNSResultMock('"v=spf1 redirect=_spf.google.com ~all"')}
}


class DNSCheckTestCase(TestCase):
    ok_domain = 'sandbox.munch.example.com'
    bad_domain = 'bad.sandbox.munch.example.com'
    inexistant_domain = 'donotexist.munch.example.com'

    def setUp(self):
        settings.DOMAINS['DKIM_KEY_ID'] = 'test'

        self.user_1 = UserFactory()
        self.user_2 = UserFactory()

        self.resolver = mock_resolver(mock_queries)
        with fake_time('2099-10-10 06:00:00'):
            with patch('dns.resolver.query', self.resolver):
                SendingDomainFactory(
                    organization=self.user_1.organization, name=self.ok_domain)
                SendingDomainFactory(
                    organization=self.user_1.organization,
                    name=self.bad_domain)
                SendingDomainFactory(
                    organization=self.user_1.organization,
                    name=self.inexistant_domain)
                SendingDomainFactory(
                    organization=self.user_2.organization, name='example.com')
                SendingDomainFactory(
                    organization=self.user_2.organization,
                    name='gentle.example.com')


@override_settings(BYPASS_DNS_CHECKS=False)
class SendingDomainDKIMTestCase(DNSCheckTestCase):
    def test_valid_domain(self):
        dom = SendingDomain.objects.get(name=self.ok_domain)
        with patch('dns.resolver.query', self.resolver):
            dom.validate_dkim()
            self.assertEqual(dom.dkim_status, DomainCheckField.OK)

    def test_inexistant_domain(self):
        dom = SendingDomain.objects.get(name=self.inexistant_domain)
        with patch('dns.resolver.query', self.resolver):
            dom.validate_dkim()
            self.assertEqual(dom.dkim_status, DomainCheckField.NOT_CONFIGURED)

    def test_invalid_domain(self):
        dom = SendingDomain.objects.get(name=self.bad_domain)
        with patch('dns.resolver.query', self.resolver):
            dom.validate_dkim()
            self.assertEqual(
                dom.dkim_status, DomainCheckField.BADLY_CONFIGURED)


@override_settings(BYPASS_DNS_CHECKS=True)
class DomainSyntaxValidation(DNSCheckTestCase):
    def test_invalid_domain(self):
        sd = SendingDomain(
            name='notadomainright', organization=self.user_2.organization)
        with self.assertRaises(ValidationError):
            sd.full_clean()

    def test_valid_domain(self):
        sd = SendingDomain(
            name='example.com', organization=self.user_2.organization)
        with self.assertRaises(ValidationError):
            sd.full_clean()


@override_settings(BYPASS_DNS_CHECKS=False)
class EnvelopeDomainValidation(DNSCheckTestCase):
    def setUp(self):
        super().setUp()
        self.v = SPFValidator(spf_include='spf.munch.example.com')

    def test_valid_domain(self):
        with patch('dns.resolver.query', self.resolver):
            self.v.validate(self.ok_domain)

    def test_inexistant_domain(self):
        with patch('dns.resolver.query', self.resolver):
            with self.assertRaisesRegexp(ValidationError, 'not configured'):
                self.v.validate(self.inexistant_domain)

    def test_invalid_domain(self):
        with patch('dns.resolver.query', self.resolver):
            with self.assertRaisesRegexp(
                    ValidationError, 'badly configured'):
                self.v.validate(self.bad_domain)


@override_settings(
    BYPASS_DNS_CHECKS=False, RETURNPATH_DOMAIN='sandbox.munch.example.com')
class SenderEmailValidation(DNSCheckTestCase):
    def setUp(self):
        super().setUp()
        settings.DOMAINS['SPF_INCLUDE'] = 'spf.munch.example.com'

    def _mk_message(self, sender):
        return MessageFactory(sender_email=sender, author=self.user_1)

    def test_savable_message(self):
        with patch('dns.resolver.query', self.resolver):
            self._mk_message('test@' + self.ok_domain)

    def test_unsavable_message_invalid_domain(self):
        with patch('dns.resolver.query', self.resolver):
            with self.assertRaises(ValidationError):
                self._mk_message('test@' + self.bad_domain)

    def test_unsavable_message_unregistered_domain(self):
        SendingDomain.objects.all().delete()
        with patch('dns.resolver.query', self.resolver):
            with self.assertRaises(ValidationError):
                self._mk_message('test@' + self.ok_domain)


@override_settings(BYPASS_DNS_CHECKS=False)
class SendingDomainTaskTestCase(DNSCheckTestCase):
    def test_untouch_domains(self):
        self.assertEqual(
            SendingDomain.objects.get(
                name='sandbox.munch.example.com').dkim_status,
            DomainCheckField.OK)
        self.assertEqual(
            SendingDomain.objects.get(
                name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.BADLY_CONFIGURED)
        self.assertEqual(
            SendingDomain.objects.get(
                name='donotexist.munch.example.com').dkim_status,
            DomainCheckField.NOT_CONFIGURED)

        with patch('dns.resolver.query', self.resolver):
            run_domains_validation([
                DomainCheckField.OK, DomainCheckField.NOT_CONFIGURED,
                DomainCheckField.BADLY_CONFIGURED, DomainCheckField.PENDING,
                DomainCheckField.UNKNOWN])

        self.assertEqual(
            SendingDomain.objects.get(
                name='sandbox.munch.example.com').dkim_status,
            DomainCheckField.OK)
        self.assertEqual(
            SendingDomain.objects.get(
                name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.BADLY_CONFIGURED)
        self.assertEqual(
            SendingDomain.objects.get(
                name='donotexist.munch.example.com').dkim_status,
            DomainCheckField.NOT_CONFIGURED)

    def test_bad_to_well(self):
        mock_queries = {
            'test._domainkey.bad.sandbox.munch.example.com': {
                'TXT': DNSResultMock(settings.DOMAINS['DKIM_KEY_CONTENT'])}
        }
        resolver = mock_resolver(mock_queries)
        with patch('dns.resolver.query', resolver):
            run_domains_validation([DomainCheckField.BADLY_CONFIGURED])

        self.assertEqual(
            SendingDomain.objects.get(
                name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.OK)

    def test_update_dkim_status_date(self):
        mock_queries = {
            'test._domainkey.bad.sandbox.munch.example.com': {
                'TXT': DNSResultMock(settings.DOMAINS['DKIM_KEY_CONTENT'])}
        }
        resolver = mock_resolver(mock_queries)

        domain = SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com')

        with fake_time('2099-10-10 08:00:00'):
            with patch('dns.resolver.query', resolver):
                validate_sending_domain_field(domain.id, 'dkim')

        self.assertEqual(SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com').dkim_status_date,
            datetime(year=2099, month=10, day=10, hour=6).replace(
                tzinfo=pytz.UTC))
        self.assertEqual(
            SendingDomain.objects.get(
                name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.OK)

    def test_ignore_old_dkim_status_change_when_bad(self):
        """
        'run_domains_validation' task must ignore
        domains that failed since more than 2 days.
        In order to prevent checking a badly configured
        domain during a long time.
        """
        mock_queries = {
            'test._domainkey.bad.sandbox.munch.example.com': {
                'TXT': DNSResultMock(settings.DOMAINS['DKIM_KEY_CONTENT'])}
        }
        resolver = mock_resolver(mock_queries)

        # Must be OK with "resolver"
        with fake_time('2099-10-12 08:00:00'):
            with patch('dns.resolver.query', resolver):
                run_domains_validation([DomainCheckField.BADLY_CONFIGURED])

        self.assertEqual(SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.BADLY_CONFIGURED)
        self.assertEqual(SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com').dkim_status_date,
            datetime(year=2099, month=10, day=10, hour=4).replace(
                tzinfo=pytz.UTC))

    def test_do_not_ignore_dkim_status_change_when_ok(self):
        """
        'run_domains_validation' task wont ignore status change for domains
        that were 'OK'. In order to always check 'OK' domains for a long time
        """
        mock_queries = {
            'test._domainkey.bad.sandbox.munch.example.com': {
                'TXT': DNSResultMock(settings.DOMAINS['DKIM_KEY_CONTENT'])}
        }
        resolver = mock_resolver(mock_queries)

        # Must be OK with "resolver"
        with fake_time('2099-10-10 08:00:00'):
            with patch('dns.resolver.query', resolver):
                run_domains_validation([DomainCheckField.BADLY_CONFIGURED])

        self.assertEqual(SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.OK)

        # Must fail with "self.resolver"
        with fake_time('2100-10-10 08:00:00'):
            with patch('dns.resolver.query', self.resolver):
                run_domains_validation([DomainCheckField.OK])

        self.assertEqual(SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com').dkim_status,
            DomainCheckField.BADLY_CONFIGURED)
        self.assertEqual(SendingDomain.objects.get(
            name='bad.sandbox.munch.example.com').dkim_status_date,
            datetime(year=2100, month=10, day=10, hour=6).replace(
                tzinfo=pytz.UTC))



import re

from django.core.validators import RegexValidator


# SMTP status codes
rfc3463_regex = re.compile(r'([234](\.\d{1,3}){2})')
rfc3463_regex_validator = RegexValidator(
    regex=rfc3463_regex.pattern,
    message='status code should follow rfc3463')



import logging
import collections
from urllib.parse import urlparse

from django.db.models import Prefetch
from django.shortcuts import get_object_or_404
from django.core.cache import cache
from django.urls import resolve
from django.urls import Resolver404
from django.urls import get_script_prefix
from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.decorators import detail_route
from rest_framework import status
from rest_framework_bulk.mixins import BulkCreateModelMixin
from rest_framework_bulk.mixins import BulkDestroyModelMixin
from rest_framework.renderers import JSONRenderer
from rest_framework.renderers import BrowsableAPIRenderer

from munch.core.utils.renderers import HTMLRenderer
from munch.core.utils.renderers import PlainTextRenderer
from munch.core.utils.views import filtered
from munch.core.utils.views import paginated
from munch.core.utils.views import NestedView
from munch.core.utils.views import classproperty
from munch.core.utils.views import MunchModelViewSetMixin
from munch.core.utils.views import OrganizationOwnedViewSetMixin
from munch.apps.optouts.models import OptOut
from munch.apps.optouts.api.v1.filters import OptOutFilter
from munch.apps.optouts.api.v1.serializers import OptOutSerializer

from ...models import Mail
from ...models import Message
from ...models import MailStatus
from ...models import PreviewMail
from ...models import MessageAttachment
from .filters import MailFilter
from .filters import MessageFilter
from .filters import MailStatusFilter
from .serializers import MailSerializer
from .serializers import MessageSerializer
from .serializers import MailStatusSerializer
from .serializers import MailBulkResultSerializer
from .serializers import NestedMailStatusSerializer
from .serializers import MessageAttachmentSerializer
from .serializers import PreviewRecipientsSerializer
from .permissions import MailPermission
from .permissions import AttachmentPermission
from .permissions import PreviewMailPermission

log = logging.getLogger(__name__)


class MessageViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = Message
    serializer_class = MessageSerializer
    filter_class = MessageFilter

    def _is_message_editable_state(self, pk):
        status = get_object_or_404(Message, pk=pk).status
        return not (status in (Message.SENDING, Message.SENT))

    def destroy(self, request, pk=None):
        if self._is_message_editable_state(pk):
            return super().destroy(request, pk)
        else:
            return Response({'error': 'Cannot delete a sending campaign'},
                            status=status.HTTP_405_METHOD_NOT_ALLOWED)

    def update(self, request, pk=None, *args, **kwargs):
        if self._is_message_editable_state(pk):
            return super().update(request, pk, *args, **kwargs)
        else:
            return Response({'error': 'Cannot update a sending campaign'},
                            status=status.HTTP_405_METHOD_NOT_ALLOWED)

    def partial_update(self, request, pk=None):
        if self._is_message_editable_state(pk):
            return super().partial_update(request, pk)
        else:
            return Response({'error': 'Cannot update a sending campaign'},
                            status=status.HTTP_405_METHOD_NOT_ALLOWED)

    @detail_route(methods=['get'])
    @paginated(MessageAttachmentSerializer)
    def attachments(self, request, pk):
        message = self.get_object()
        return message.attachments.all()

    @detail_route(methods=['get'])
    @paginated(MailStatusSerializer)
    @filtered(MailStatusFilter)
    def bounces(self, request, pk):
        message = self.get_object()
        return MailStatus.objects.filter(
            status__in=(MailStatus.BOUNCED, MailStatus.DROPPED),
            mail__message=message)

    @detail_route(methods=['get'])
    @paginated(MailSerializer)
    @filtered(MailFilter)
    def recipients(self, request, pk):
        return self.get_object().mails.all()

    @detail_route(methods=['get'])
    @paginated(OptOutSerializer)
    @filtered(OptOutFilter)
    def opt_outs(self, request, pk, format=None):
        identifiers = Mail.objects.filter(
            message=self.get_object()).values_list('identifier', flat=True)
        return OptOut.objects.filter(identifier__in=identifiers)

    @detail_route(
        methods=['get'], renderer_classes=(
            JSONRenderer, BrowsableAPIRenderer,
            HTMLRenderer, PlainTextRenderer))
    def preview(self, request, pk=None, format=None):
        msg = self.get_object()
        if format == 'txt':
            resp = msg.mk_plaintext()
        elif format == 'html':
            resp = msg.mk_html()
        else:
            resp = collections.OrderedDict((
                ('recipients', msg.willsend_addresses()),
                ('excluded_recipients', msg.willnotsend_addresses()),
                ('spam_score', msg.spam_score),
                ('is_spam', msg.is_spam),
                ('html', msg.mk_html()),
                ('plaintext', msg.mk_plaintext()),
            ))
        return Response(resp)

    @detail_route(methods=['get'])
    def stats(self, request, pk, format=None):
        message = self.get_object()
        return Response(message.mk_stats())


class PreviewSendMessageView(NestedView):
    """ Sends a preview email to some recipients

    Not comptabilized in stats.
    """
    serializer_class = PreviewRecipientsSerializer
    parent_model = Message
    permission_classes = [PreviewMailPermission]

    def post(self, request, msg_pk, format=None):
        m = self.get_parent_object(msg_pk)

        serializer = self.get_serializer(data=request.data)

        if serializer.is_valid():
            addrs = serializer.validated_data['recipient']
            mails = [
                PreviewMail.objects.create(
                    message=m, recipient=i) for i in addrs]
            [PreviewMail.send_preview(mail.pk) for mail in mails]
            return Response('Preview mail sent to {}'.format(addrs))
        else:
            return Response(
                serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class MailViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        BulkCreateModelMixin,
        BulkDestroyModelMixin,
        viewsets.ModelViewSet):
    model = Mail
    serializer_class = MailSerializer
    filter_class = MailFilter
    permission_classes = [MailPermission]

    def get_queryset(self, with_statuses=True):
        qs = super().get_queryset()
        if with_statuses:
            status_qs = MailStatus.objects.order_by(
                'mail', '-creation_date').distinct('mail')
            qs = qs.prefetch_related(Prefetch(
                'statuses', queryset=status_qs, to_attr='last_status_cached'))
        return qs

    def bulk_destroy(self, request, *args, **kwargs):
        if not request.data:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        qs = self.get_queryset()
        ids = []

        for url in request.data:
            try:
                http_prefix = url.startswith(('http:', 'https:'))
            except AttributeError:
                return Response(status=status.HTTP_400_BAD_REQUEST)

            if http_prefix:
                # If needed convert absolute URLs to relative path
                url = urlparse(url).path
                prefix = get_script_prefix()
                if url.startswith(prefix):
                    url = '/' + url[len(prefix):]

            try:
                match = resolve(url)
            except Resolver404:
                return Response(status=status.HTTP_400_BAD_REQUEST)

            try:
                pk = int(match.kwargs.get('pk'))
            except (ValueError, TypeError):
                return Response(status=status.HTTP_400_BAD_REQUEST)

            ids.append(pk)

        if not ids:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        qs.filter(id__in=ids).delete()

        return Response(status=status.HTTP_204_NO_CONTENT)

    def create(self, request, *args, **kwargs):
        bulk = isinstance(request.data, list)

        must_lock = False
        message_id = None

        if request.data:
            if bulk:
                if request.data and isinstance(request.data[0], dict):
                    message_id = request.data[0].get('message')
            else:
                message_id = request.data.get('message')

            if message_id:
                must_lock = True

        if must_lock:
            try:
                match = resolve(urlparse(message_id).path)
                message_id = match.kwargs.get('pk')
            except Resolver404:
                message_id = None

            if message_id:
                lock_name = 'mail-lock:{}'.format(message_id)
                log.debug('[{}] Waiting for lock ({})...'.format(
                    match.view_name, lock_name))
                lock = cache.lock(lock_name, timeout=15)
                lock.acquire()

        try:
            if not bulk:
                return super(BulkCreateModelMixin, self).create(
                    request, *args, **kwargs)
            else:

                serializer = self.get_serializer(data=request.data, many=True)
                serializer.build_detailed_response()

                recipients = []
                for item in serializer._validated_data:
                    recipients.append(item['recipient'])

                self.perform_bulk_create(serializer)

                mails = self.get_queryset(
                    with_statuses=False).filter(
                        recipient__in=recipients).filter(
                            message=message_id)
                mails_serializer = MailBulkResultSerializer(
                    mails, many=True, context={'request': request})

                detailed_response = serializer._detailed_response
                detailed_response['results'] = mails_serializer.data

                try:
                    lock.release()
                except:
                    pass

                return Response(
                    data=detailed_response,
                    status=status.HTTP_201_CREATED)
        except Exception:
            try:
                lock.release()
            except:
                pass

            raise

    @detail_route(methods=['get'])
    @paginated(NestedMailStatusSerializer)
    @filtered(MailStatusFilter)
    def status_log(self, request, pk):
        mail = self.get_object()
        return mail.statuses.order_by('creation_date')


class MailOptOutView(NestedView):
    """ OptOuts the selected mail, via API
    """
    parent_model = Mail

    def post(self, request, mail_pk, format=None):
        mail = self.get_parent_object(mail_pk)
        OptOut.objects.create(
            author=mail.message.author,
            category=mail.message.category,
            identifier=mail.identifier,
            address=mail.recipient,
            origin=OptOut.BY_API)
        return Response()


class BouncesViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = MailStatus
    serializer_class = MailStatusSerializer
    filter_class = MailStatusFilter

    def get_queryset(self):
        return super().get_queryset().bounces()


class MessageAttachmentViewset(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = MessageAttachment
    serializer_class = MessageAttachmentSerializer

    @classproperty
    @classmethod
    def permission_classes(cls):
        return cls.mk_permission_classes([AttachmentPermission])

    def _get_file_response(self, attachment):
        from django.http import HttpResponse
        filename, content, mimetype = attachment.to_email_attachment()
        response = HttpResponse(content)
        response['Content-Type'] = mimetype
        response['Content-Length'] = attachment.size
        return (response, filename)

    @detail_route(methods=['get'])
    def download(self, request, pk, format=None):
        response, filename = self._get_file_response(self.get_object())
        response['Content-Disposition'] = 'attachment; filename={}'.format(
            filename)
        return response

    @detail_route(methods=['get'])
    def content(self, request, pk, format=None):
        response, filename = self._get_file_response(self.get_object())
        response['Content-Disposition'] = 'inline; filename={}'.format(
            filename)
        return response



from django.core.management.base import BaseCommand

from ...models import TrackRecord


class Command(BaseCommand):
    help = "Build tracking cache"

    def handle(self, *args, **options):
        count = TrackRecord.clear_cache()
        self.stdout.write('TrackingRecord deleted: {}'.format(count))
        count = 0
        for track_record in TrackRecord.objects.all():
            count += track_record.cache()
        self.stdout.write('TrackingRecord cached: {}'.format(count))



from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions
from rest_framework.versioning import NamespaceVersioning as BaseVersioning


class NamespaceVersioning(BaseVersioning):
    """
    Temporary class
    https://github.com/tomchristie/django-rest-framework/pull/4010
    """

    invalid_version_message = _('Invalid version in namespace.')

    def determine_version(self, request, *args, **kwargs):
        resolver_match = getattr(request, 'resolver_match', None)
        if (resolver_match is None or not resolver_match.namespace):
            return self.default_version
        version = resolver_match.namespace.split(':')[0]
        if not self.is_allowed_version(version):
            raise exceptions.NotFound(self.invalid_version_message)
        return version



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-28 12:15
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0005_auto_20160926_1810'),
    ]

    operations = [
        migrations.AddField(
            model_name='munchuser',
            name='invited_by',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='invited by'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-29 14:27
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0006_munchuser_invited_by'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='organizationsettings',
            name='notification_email',
        ),
        migrations.AlterField(
            model_name='organization',
            name='contact_email',
            field=models.EmailField(help_text='Main contact email for organization. Will also receive status/optouts notifications', max_length=254, verbose_name='contact email address'),
        ),
    ]



import os

import click
from django.utils.module_loading import import_string

import munch


@click.group()
@click.option(
    '--config',
    default='',
    envvar='DJANGO_SETTINGS_MODULE',
    help='Path to settings module.',
    metavar='PATH')
@click.version_option(version=munch.__version__)
@click.pass_context
def cli(ctx, config):
    """Munch is an emailing platform.

    Default settings module is `munch.settings` but
    it can be overridden with `DJANGO_CONFIG_MODULE`
    or with `--config` parameter.
    """
    # Elevate --config option to DJANGO_CONFIG_MODULE env var
    if config:
        os.environ['DJANGO_SETTINGS_MODULE'] = config
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'munch.settings')


list(map(lambda cmd: cli.add_command(import_string(cmd)), (
    'munch.runner.commands.run.run',
    'munch.runner.commands.help.help',
    'munch.runner.commands.django.django',
)))


def make_django_command(name, django_command=None, help=None):
    "A wrapper to convert a Django subcommand a Click command"
    if django_command is None:
        django_command = name

    @click.command(
        name=name,
        help=help,
        add_help_option=False,
        context_settings=dict(
            ignore_unknown_options=True,
        ))
    @click.argument('management_args', nargs=-1, type=click.UNPROCESSED)
    @click.pass_context
    def inner(ctx, management_args):
        from munch.runner.commands.django import django
        ctx.params['management_args'] = (django_command,) + management_args
        ctx.forward(django)

    return inner


list(map(cli.add_command, (
    make_django_command('migrate', help=(
        'Run migrations (like `munch django migrate`).')),
    make_django_command('shell', help='Run a Python interactive interpreter.')
)))


def main():
    cli(obj={}, max_content_width=100)



from django.conf import settings
from django.utils.translation import ugettext as _
from rest_framework import serializers
from rest_framework import fields
from rest_framework.settings import api_settings
from rest_framework.serializers import ValidationError
from rest_framework.serializers import DjangoValidationError
from rest_framework.serializers import get_validation_error_detail

from munch.core.models import Category
from munch.core.utils.serializers import HALLinksField
from munch.core.utils.serializers import SizeLimitedListSerializer
from munch.core.utils.serializers import ForwardValidationErrorsMixin
from munch.core.utils.relations import ScopedHyperLinkedRelatedField

from ...fields import EmailListField
from ...validators import validate_no_virus
from ...models import Mail
from ...models import Message
from ...models import MailStatus
from ...models import MessageAttachment


class NestedTrackingSummarySerializer(serializers.Serializer):
    opened = serializers.DateTimeField(source='first_open')
    open_time = serializers.IntegerField()
    clicked = serializers.ListField(
        source='clicks', child=serializers.DictField())


class NestedMailStatusSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = MailStatus
        fields = ('status', 'creation_date', 'raw_msg')


class CachedHyperlinkedRelatedField(serializers.HyperlinkedRelatedField):
    """ HyperlinkedRelatedField but with a cache on the deserializing part

    Cache is only valid the time of the Field lifetime (usually dies with the
    request).
    """
    def __init__(self, *args, **kwargs):
        self._memoize_cache = {}
        super().__init__(*args, **kwargs)

    def get_object(self, view_name, view_args, view_kwargs):
        cache = self._memoize_cache
        key = (view_name, tuple(view_kwargs.items()))

        v = cache.get(key, None)

        if not v:
            v = super().get_object(view_name, view_args, view_kwargs)
            cache[key] = v

        # If no cached version : store and return
        return v


class CachedScopedHyperLinkedRelatedField(
        CachedHyperlinkedRelatedField,
        ScopedHyperLinkedRelatedField):
    pass


########
# Mail #
########
class MailBulkSerializer(serializers.HyperlinkedModelSerializer):
    """
    This serializer is only used to validate bulk item for MailListSerializer
    """
    message = CachedScopedHyperLinkedRelatedField(
        view_name='campaigns:message-detail',
        queryset=Message.objects.all())
    # TODO: Apiv2 rename
    to = serializers.EmailField(source='recipient')

    class Meta:
        model = Mail
        fields = ['to', 'message', 'source_type', 'source_ref', 'properties']

    def run_validation(self, data=[]):
        """
        We override the default `run_validation`, because the validation
        performed by validators and the `.validate()` method should
        be coerced into an error dictionary with a 'non_fields_error' key.
        """
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data

        try:
            value = self.to_internal_value(data)
            self.run_validators(value)
            value = self.validate(value)
            assert value is not None, '.validate() should return the validated data'  # noqa
        except (ValidationError, DjangoValidationError) as exc:
            # TODO: Must be 'recipient' instead of 'to' in v2
            raise ValidationError(
                detail={
                    'to': data['to'],
                    'errors': get_validation_error_detail(exc)})

        return value

    def get_validators(self):
        """
        Determine the set of validators to use when instantiating serializer.
        """
        # If the validators have been declared explicitly then use that.
        validators = getattr(getattr(self, 'Meta', None), 'validators', None)
        if validators is not None:
            return validators[:]

        # Otherwise use the default set of validators.
        return self.get_unique_for_date_validators()


class MailListSerializer(SizeLimitedListSerializer):
    class Meta:
        # make it lazy, otherwise, overrided settings value can be missed
        max_items = lambda: settings.CAMPAIGNS['MAX_BULK_EMAILS']

    def create(self, validated_data):
        mails = [Mail(**item) for item in validated_data]
        return Mail.objects.bulk_create(mails, batch_size=30000)

    def build_detailed_response(self):
        assert hasattr(self, 'initial_data'), (
            'Cannot call `.is_partly_valid()`no `data=` keyword argument was'
            'passed when instantiating the serializer instance.'
        )
        # this one may trigger a ValidationError up to http client
        self.run_list_size_validation(self.initial_data)

        # those can't, from here, validation is always considered good
        if not hasattr(self, '_validated_data'):
            self._validated_data, detail = self.run_partial_validation(
                self.initial_data)
            self._detailed_response = detail

    def run_partial_validation(self, data=fields.empty):
        """
        We override the default `run_validation`, because the validation
        performed by validators and the `.validate()` method should
        be coerced into an error dictionary with a 'non_fields_error' key.
        """
        # Called 1 time for all the data list
        # output of this function is set into self._validated_data by is_valid
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data
        value, detail = self.to_internal_value_error(data)
        try:
            self.run_validators(value)  # does nothing in our case
            value = self.validate(value)
            assert value is not None, \
                '.validate() should return the validated data'
        except (ValidationError, DjangoValidationError) as exc:
            raise ValidationError(detail=get_validation_error_detail(exc))
        return value, detail

    def to_internal_value_error(self, data):
        """ Returns both errors and valid data
        :returns: a couple: list of valid values
                            ValidationError with all errors
        """
        # Called 1 time for all the data list

        if fields.html.is_html_input(data):
            data = fields.html.parse_html_list(data)

        if not isinstance(data, list):
            message = self.error_messages['not_a_list'].format(
                input_type=type(data).__name__
            )
            raise ValidationError({
                api_settings.NON_FIELD_ERRORS_KEY: [message]
            })

        detail = {
            'validation_errors': {},
            'no_address': 0,
            'duplicates': [],
            'results': []
        }
        unique_couples = set()
        messages = []

        cleaned_data = []
        for item in data:
            if not isinstance(item, dict):
                detail['no_address'] += 1
                continue
            # TODO: Must be 'recipient' instead of 'to' in v2
            if not item.get('to') or not item.get('message'):
                detail['no_address'] += 1
                continue

            # TODO: Must be 'recipient' instead of 'to' in v2
            unicity = (item['message'], item['to'])
            if unicity not in unique_couples:
                unique_couples.add(unicity)
            else:
                # TODO: Must be 'recipient' instead of 'to' in v2
                detail['duplicates'].append(item['to'])
                continue

            if item['message'] not in messages:
                if len(messages) > 0:
                    detail[api_settings.NON_FIELD_ERRORS_KEY] = [
                        'Every object must have same message']
                    return [], detail
                messages.append(item['message'])

            cleaned_data.append(item)

        serializer = MailBulkSerializer(
            data=cleaned_data, many=True, context=self.context)

        if not serializer.is_valid():
            in_error, only_valid = [], []

            for error in serializer.errors:
                if error:
                    in_error.append(error['to'])
                    detail['validation_errors'][error['to']] = error['errors']

            for item in cleaned_data:
                if item['to'] not in in_error:
                    only_valid.append(item)

            serializer = MailBulkSerializer(
                data=only_valid, many=True, context=self.context)
            serializer.is_valid()

        cleaned_data = serializer.validated_data

        if cleaned_data:
            recipients = []
            results = []
            for item in cleaned_data:
                recipients.append(item['recipient'])

            recipients_to_remove = Mail.objects.filter(
                message=cleaned_data[0]['message'],
                recipient__in=recipients).values_list('recipient', flat=True)

            for item in cleaned_data:
                if item['recipient'] not in recipients_to_remove:
                    results.append(item)
                else:
                    detail['validation_errors'][item['recipient']] = [
                        _('This address is already attached to this message.')]

            cleaned_data = results

        return cleaned_data, detail


class MailBulkResultSerializer(serializers.HyperlinkedModelSerializer):
    to = serializers.EmailField(source='recipient')
    message = CachedScopedHyperLinkedRelatedField(
        view_name='campaigns:message-detail',
        queryset=Message.objects.all())
    # TODO: Apiv2 rename
    date = serializers.DateTimeField(source='creation_date', read_only=True)

    class Meta:
        model = Mail
        fields = [
            'url', 'to', 'date', 'message',
            'source_type', 'source_ref', 'properties']
        extra_kwargs = {'url': {'view_name': 'campaigns:recipient-detail'}}


class MailSerializer(serializers.HyperlinkedModelSerializer):
    last_status = NestedMailStatusSerializer(read_only=True)
    tracking = NestedTrackingSummarySerializer(
        read_only=True, source='tracking_info')
    message = CachedScopedHyperLinkedRelatedField(
        view_name='campaigns:message-detail',
        queryset=Message.objects.all())
    # TODO: Apiv2 rename
    to = serializers.EmailField(source='recipient')
    _links = HALLinksField(
        nested_endpoints=['optout', 'status_log'],
        view_name='campaigns:recipient-detail')
    # TODO: Apiv2 rename
    date = serializers.DateTimeField(source='creation_date', read_only=True)
    delivery_status = serializers.CharField(source='curstatus', read_only=True)

    class Meta:
        model = Mail
        fields = [
            'url', 'to', 'date', 'last_status', 'message', 'tracking',
            'delivery_status', 'source_type', 'source_ref', 'properties',
            '_links']
        read_only_fields = ['date', 'delivery_status']
        list_serializer_class = MailListSerializer
        extra_kwargs = {'url': {'view_name': 'campaigns:recipient-detail'}}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        extra_fields = ['last_status', 'tracking']

        request = kwargs['context'].get('request')
        if request and request.method == 'GET':
            for field in request.GET.get('extra_fields', '').split(','):
                if field in extra_fields:
                    extra_fields.remove(field)
            for field in extra_fields:
                self.fields.pop(field)

    def validate_message(self, value):
        if value.status in [Message.SENDING, Message.SENT]:
            raise ValidationError(_(
                "Cannot add this recipient to a "
                "sending or already sent message."))
        return value


###########
# Message #
###########
class MessageSerializer(
        ForwardValidationErrorsMixin,
        serializers.HyperlinkedModelSerializer):
    completion_date = serializers.ReadOnlyField()
    creation_date = serializers.ReadOnlyField()
    msg_issue = serializers.ReadOnlyField()
    is_spam = serializers.ReadOnlyField()
    spam_score = serializers.ReadOnlyField()
    recipient_count = serializers.ReadOnlyField(source='mails.count')
    category = ScopedHyperLinkedRelatedField(
        view_name='core:category-detail',
        queryset=Category.objects.all(),
        required=False, allow_null=True)

    _links = HALLinksField(
        nested_endpoints=[
            'preview_send', 'preview', 'preview.html', 'preview.txt',
            'recipients', 'bounces', 'opt_outs', 'stats', 'attachments'],
        endpoints={'archive_url': {
            'view_name': 'message_web_view',
            'lookup_field': 'identifier',
            'lookup_key': 'identifier'}},
        view_name='campaigns:message-detail')

    def validate(self, attrs):
        """
        http://www.django-rest-framework.org/topics/3.0-announcement/#differences-between-modelserializer-validation-and-modelform
        """
        request = self.context.get('request', None)
        # Re-use model validation logic
        instance = Message(author=request.user, **attrs)
        try:
            instance.clean()
            if 'html' in attrs:
                instance.validate_html()
        except DjangoValidationError as err:
            message = {}
            for field, errors in err.message_dict.items():
                if not isinstance(errors, list):
                    errors = [errors]
                message[field] = errors
            raise ValidationError(message)
        return attrs

    class Meta:
        model = Message
        read_only_fields = ['send_date']
        fields = (
            'url', 'id', 'name', 'sender_email', 'sender_name', 'subject',
            'html', 'status', 'category', 'recipient_count', 'properties',
            'creation_date', 'send_date', 'completion_date', 'track_open',
            'track_clicks', 'external_optout', 'detach_images', 'spam_score',
            'spam_details', 'is_spam', 'msg_issue', '_links', )
        extra_kwargs = {'url': {'view_name': 'campaigns:message-detail'}}


class MailStatusSerializer(serializers.HyperlinkedModelSerializer):
    address = serializers.ReadOnlyField(source='mail.recipient')

    class Meta:
        model = MailStatus
        fields = ('status', 'creation_date', 'raw_msg', 'address')


class MessageAttachmentSerializer(serializers.HyperlinkedModelSerializer):
    file = serializers.FileField(
        validators=[validate_no_virus], write_only=True)

    _links = HALLinksField(
        nested_endpoints=['download', 'content'],
        view_name='campaigns:messageattachment-detail')

    def validate(self, attrs):
        # this is defined as a pre_save because cant be checked in clean() as
        # there is not yet instance.message at this step.
        same_named = attrs['message'].attachments.filter(
            file__endswith='/{}'.format(attrs['file'].name))
        if same_named.exists():
            raise ValidationError((
                'A file with the same name already exists '
                'for this message : {}').format(same_named.first()))

        # Now check whether we would fit within the total allowed
        # attachment size for the message
        total_size = 0
        for attachment in attrs['message'].attachments.all():
            total_size += attachment.size
        total_size += attrs['file'].size
        if total_size > settings.CAMPAIGNS['ATTACHMENT_TOTAL_MAX_SIZE']:
            raise ValidationError(_(
                'Can not accept this file. Total attachment size for this '
                'message ({} bytes) would exceed the maximum allowed size of '
                '{} bytes'.format(
                    total_size, settings.CAMPAIGNS[
                        'ATTACHMENT_TOTAL_MAX_SIZE'])))
        return attrs

    def validate_file(self, value):
        # Check this attachment's size to make sure it's not too big
        if value.size > settings.CAMPAIGNS['ATTACHMENT_MAX_SIZE']:
            raise ValidationError(_(
                'This file is too big. Maximum allowed size is '
                '{} bytes'.format(settings.CAMPAIGNS['ATTACHMENT_MAX_SIZE'])))
        if len(value.name) > 100:
            raise ValidationError(_(
                'This file name is too long. It must be under 100 characters'))
        return value

    def create(self, validated_data):
        validated_data.update({'original_name': validated_data['file'].name})
        return super().create(validated_data)

    class Meta:
        model = MessageAttachment
        fields = (
            'url', 'message', 'file', 'filename', 'size', 'size_in_mail',
            '_links')
        readonly_fields = ('filename', 'size', 'size_in_mail')
        extra_kwargs = {
            'url': {'view_name': 'campaigns:messageattachment-detail'},
            'message': {'view_name': 'campaigns:message-detail'}}


class EmailListSerializer(serializers.CharField):
    """ Serializer for comma-separated emails in a string

    Re-use EmailListField logic.
    """
    default_validators = [EmailListField.EmailListValidator()]

    def to_representation(self, obj):
        f = EmailListField()
        char_obj = super().to_representation(obj)
        return f.value_to_string(char_obj)

    def to_internal_value(self, data):
        f = EmailListField()
        char_str = super().to_internal_value(data)
        return f.to_python(char_str)


class PreviewRecipientsSerializer(serializers.Serializer):
    to = EmailListSerializer(source='recipient')



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-05-27 08:26
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import munch.apps.upload_store.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('users', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Image',
            fields=[
                ('upload_date', models.DateTimeField(auto_now=True)),
                ('expiration', models.DurationField()),
                ('hash', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('file', models.ImageField(upload_to=munch.apps.upload_store.models.Upload.get_path)),
                ('width', models.PositiveSmallIntegerField(default=600, validators=[django.core.validators.MaxValueValidator(600)])),
                ('organization', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images', to='users.Organization')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
